from .ICF import ICF
import numpy as np
from tqdm import tqdm
from threadpoolctl import threadpool_limits
import ctypes
from numba import jit
import scipy
import mf
import joblib


@jit(nopython=True)
def _central_limit_theorem(k):
    p = len(k)
    x = (np.sum(k) - p / 2) / (np.sqrt(p / 12))
    return x


@jit(nopython=True)
def _numba_multivariate_normal(mean, cov):
    n = len(mean)
    cov_eig = np.linalg.eigh(cov)  # suppose that the matrix is symmetric
    x = np.zeros(n)
    for i in range(n):
        x[i] = _central_limit_theorem(
            np.random.uniform(0, 1, 200)
        )  # best parameter is 20000 in terms of speed and accuracy in distribution sampling
    return ((np.diag(cov_eig[0])**(0.5)) @ cov_eig[1].T @ x) + mean


@jit(nopython=True)
def _sample_items_weights(user_candidate_items, items_means, items_covs):
    n = len(user_candidate_items)
    num_lat = items_means.shape[1]
    qs = np.zeros((n, num_lat))
    for i, item in enumerate(user_candidate_items):
        item_mean = items_means[item]
        item_cov = items_covs[item]
        qs[i] = _numba_multivariate_normal(item_mean, item_cov)
    return qs


class LinearThompsonSampling(ICF):
    """Linear Thompson Sampling.
    An adaptation of the original Thompson Sampling to measure the latent dimensions by a PMF formulation [1]_.

    References
    ----------
    .. [1] Abeille, Marc, and Alessandro Lazaric. "Linear thompson sampling revisited." 
       Artificial Intelligence and Statistics. PMLR, 2017.
    """
    def __init__(self, *args, **kwargs):
        """__init__.

        Args:
            args:
            kwargs:
        """
        super().__init__(*args, **kwargs)

    def reset(self, observation):
        """reset.

        Args:
            observation: 
        """ 
        train_dataset = observation
        super().reset(train_dataset)
        self.train_dataset = train_dataset
        self.train_consumption_matrix = scipy.sparse.csr_matrix(
            (self.train_dataset.data[:, 2],
             (self.train_dataset.data[:, 0], self.train_dataset.data[:, 1])),
            (self.train_dataset.num_total_users,
             self.train_dataset.num_total_items))
        self.num_total_items = self.train_dataset.num_total_items

        mf_model = mf.ICFPMFS(self.iterations,
                              self.var,
                              self.user_var,
                              self.item_var,
                              self.stop_criteria,
                              num_lat=self.num_lat)
        mf_model_id = joblib.hash(
            (mf_model.get_id(), self.train_consumption_matrix))

        if pdm.file_exists(mf_model_id):
            mf_model = pdm.load(mf_model_id)
        else:
            mf_model.fit(self.train_consumption_matrix)
            pdm.save(mf_model_id, mf_model)

        self.items_means = mf_model.items_means
        self.items_covs = mf_model.items_covs
        self.num_latent_factors = len(self.items_latent_factors[0])

        self.I = np.eye(self.num_latent_factors)

        bs = defaultdict(lambda: np.zeros(self.num_latent_factors))
        As = defaultdict(lambda: self.get_user_lambda() * I)

    def action_estimates(self, candidate_actions):
        """action_estimates.

        Args:
            candidate_actions: (user id, candidate_items)
        
        Returns:
            numpy.ndarray:
        """
        uid = candidate_actions[0]
        candidate_items = candidate_actions[1]
        b = bs[uid]
        A = As[uid]

        mean = np.dot(np.linalg.inv(A), b)
        cov = np.linalg.inv(A) * self.var
        p = np.random.multivariate_normal(mean, cov)
        qs = _sample_items_weights(candidate_items, self.items_means,
                                   self.items_covs)

        items_score = p @ qs.T
        return items_score, {'qs': qs, 'candidate_items': candidate_items}

    def update(self, observation, action, reward, info):
        """update.

        Args:
            observation:
            action: (user id, item)
            reward (float): reward
            info: 
        """
        uid = action[0]
        item = action[1]
        additional_data = info
        max_q = additional_data['qs'][np.argmax(
            item == additional_data['candidate_items']), :]
        A += max_q[:, None].dot(max_q[None, :])
        b += reward * max_q



from . import Recommender
import numpy as np
import mf
import ctypes
from threadpoolctl import threadpool_limits
import util


class SVDPlusPlus(Recommender):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        pass

    def train(self, train_matrix):
        super().train()
        self.train_matrix = train_matrix
        mf_model = mf.SVDPlusPlus()
        mf_model.fit(train_matrix)
        self.b_u, self.b_i, self.p, self.q, self.y, self.r_mean =\
            mf_model.b_u,mf_model.b_i,\
            mf_model.p,mf_model.q,\
            mf_model.y,mf_model.r_mean

    def predict(self, test_matrix):
        super().predict()
        test_uids = np.nonzero(np.sum(test_matrix > 0, axis=1).A.flatten())[0]
        self_id = id(self)

        with threadpool_limits(limits=1, user_api='blas'):
            args = [(
                self_id,
                int(uid),
            ) for uid in test_uids]
            results = util.run_parallel(self.predict_user_items, args)

        for i, user_result in enumerate(results):
            self.results[test_uids[i]] = user_result

        self.save_results()

    @staticmethod
    def predict_user_items(obj_id, uid):
        self = ctypes.cast(obj_id, ctypes.py_object).value
        user_consumption = self.train_matrix[uid, :].A.flatten()
        iids = np.nonzero(user_consumption)[0]
        test_iids = np.where(user_consumption == 0)[0]
        p_u = self.p[uid] + np.sum(self.y[iids], axis=0)
        items_scores = [
            self.r_mean + self.b_u[uid] + self.b_i[iid] + p_u @ self.q[iid]
            for iid in test_iids
        ]
        top_iids = test_iids[np.argsort(items_scores)
                             [::-1]][:self.result_list_size]
        return top_iids

    def filter_parameters(self, parameters):
        return super().filter_parameters(
            {k: v
             for k, v in parameters.items() if k not in ['r_mean']})



from collections import defaultdict
from tabulate import tabulate
from scipy.sparse import base

from threadpoolctl import threadpool_limits
import scipy.sparse
import numpy as np
import random
from tqdm import tqdm
import irec.value_functions
from irec.CandidateActions import OneUserCandidateActions
from irec.utils.dataset import Dataset
from irec.agents import Agent

from irec.value_functions.Entropy import Entropy
from irec.value_functions.MostPopular import MostPopular
from irec.value_functions.LogPopEnt import LogPopEnt
from irec.value_functions.BestRated import BestRated

import matplotlib as mpl
import seaborn as sns
import matplotlib.pyplot as plt
import scipy.stats
import os

from concurrent.futures import ProcessPoolExecutor
import multiprocessing
import itertools
import ctypes

"""Evaluation Policies.

This module implements several assessment policies that will define how 
to assess the agent after the recommendation process.
"""

class EvaluationPolicy:
    """EvaluationPolicy.
        
    Defines a form of evaluation for the recommendation process.
    """

    def evaluate(
        self, model: Agent, train_dataset: Dataset, test_dataset: Dataset
    ) -> [list, dict]:
        """evaluate.
        
        Performs the form of evaluation according to the chosen policy.

        Args:
            model (Agent): model
            train_dataset (Dataset): train_dataset
            test_dataset (Dataset): test_dataset

        Returns:
            Tuple[list, dict]:
        """
        pass


class Interaction(EvaluationPolicy):
    def __init__(
        self,
        num_interactions: int,
        interaction_size: int,
        save_info: bool,
        *args,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        self.num_interactions = int(num_interactions)
        self.interaction_size = int(interaction_size)
        self.save_info = save_info

    def evaluate(self, model, train_dataset, test_dataset):
        with threadpool_limits(limits=1, user_api="blas"):
            test_users = np.unique(test_dataset.data[:, 0]).astype(int)
            num_total_items = test_dataset.num_total_items
            test_consumption_matrix = scipy.sparse.csr_matrix(
                (
                    test_dataset.data[:, 2],
                    (
                        test_dataset.data[:, 0].astype(int),
                        test_dataset.data[:, 1].astype(int),
                    ),
                ),
                shape=(test_dataset.num_total_users, test_dataset.num_total_items),
            )

            users_items_recommended = defaultdict(list)

            for i in range(len(train_dataset.data)):
                uid = int(train_dataset.data[i, 0])
                if uid in test_users:
                    iid = int(train_dataset.data[i, 1])
                    reward = train_dataset.data[i, 2]
                    users_items_recommended[uid].append(iid)

            num_test_users = len(test_users)
            print(f"Starting {model.name} Training")
            model.reset(train_dataset)
            print(f"Ended {model.name} Training")
            users_num_interactions = defaultdict(int)
            available_users = set(test_users)

            history_items_recommended = []

            num_trials = num_test_users * self.num_interactions
            # _intervals = num_trials // 20
            _intervals = num_trials // num_trials
            _num_interactions = 0
            pbar = tqdm(total=num_trials)
            pbar.set_description(f"{model.name}")
            acts_info = []
            for i in range(num_trials):
                uid = random.sample(available_users, k=1)[0]
                not_recommended = np.ones(num_total_items, dtype=bool)
                not_recommended[users_items_recommended[uid]] = 0
                items_not_recommended = np.nonzero(not_recommended)[0]
                # items_score, info = model.action_estimates((uid,items_not_recommended))
                # best_items = items_not_recommended[np.argpartition(items_score,-self.interaction_size)[-self.interaction_size:]]

                actions, info = model.act(
                    OneUserCandidateActions(uid, items_not_recommended),
                    self.interaction_size,
                )
                if self.save_info:
                    info["trial"] = i
                    info["user_interaction"] = users_num_interactions[uid]
                acts_info.append(info)
                best_items = actions[1]
                users_items_recommended[uid].extend(best_items)

                for item in best_items:
                    history_items_recommended.append((uid, item))
                    model.observe(
                        None, (uid, item), test_consumption_matrix[uid, item], info
                    )
                # model.increment_time()
                users_num_interactions[uid] += 1
                if users_num_interactions[uid] == self.num_interactions:
                    available_users = available_users - {uid}

                _num_interactions += 1
                if i % _intervals == 0 and i != 0:
                    pbar.update(_num_interactions)
                    _num_interactions = 0
            pbar.update(_num_interactions)
            _num_interactions = 0
            pbar.close()
            return history_items_recommended, acts_info


class InteractionSample(EvaluationPolicy):
    def __init__(self, num_interactions, interaction_size, rseed, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.num_interactions = num_interactions
        self.interaction_size = interaction_size
        self.rseed = rseed

    def evaluate(self, model, train_dataset, test_dataset):
        np.random.seed(self.rseed)
        test_users = np.unique(test_dataset.data[:, 0]).astype(int)
        num_total_items = test_dataset.num_total_items
        test_consumption_matrix = scipy.sparse.csr_matrix(
            (
                test_dataset.data[:, 2],
                (
                    test_dataset.data[:, 0].astype(int),
                    test_dataset.data[:, 1].astype(int),
                ),
            ),
            shape=(test_dataset.num_total_users, test_dataset.num_total_items),
        )

        data = np.vstack((train_dataset.data, test_dataset.data))
        from irec.utils.dataset import Dataset

        dataset = Dataset(data)
        dataset.update_from_data()
        dataset.update_num_total_users_items()

        consumption_matrix = scipy.sparse.csr_matrix(
            (dataset.data[:, 2], (dataset.data[:, 0], dataset.data[:, 1])),
            (dataset.num_total_users, dataset.num_total_items),
        )

        num_users_to_sample = 6
        num_consumed = (test_consumption_matrix > 0).sum(axis=1).A.flatten()
        users_selected = []
        num_users_to_sample -= len(users_selected)
        uids = np.nonzero(num_consumed >= 100)[0]
        users_sampled = np.random.choice(uids, num_users_to_sample, replace=False)
        users_selected.extend(users_sampled)

        users_items_recommended = defaultdict(list)
        num_users_selected = len(users_selected)
        print(f"Starting {model.name} Training")
        model.reset(train_dataset)
        print(f"Ended {model.name} Training")
        users_num_interactions = defaultdict(int)
        available_users = set(users_selected)

        history_items_recommended = []

        num_trials = num_users_selected * self.num_interactions
        _intervals = num_trials // 20
        _num_interactions = 0
        pbar = tqdm(total=num_trials)
        pbar.set_description(f"{model.name}")
        for i in range(num_trials):
            uid = random.sample(available_users, k=1)[0]
            not_recommended = np.ones(num_total_items, dtype=bool)
            not_recommended[users_items_recommended[uid]] = 0
            items_not_recommended = np.nonzero(not_recommended)[0]
            items_score, info = model.action_estimates((uid, items_not_recommended))
            best_items = items_not_recommended[
                np.argpartition(items_score, -self.interaction_size)[
                    -self.interaction_size :
                ]
            ]
            users_items_recommended[uid].extend(best_items)

            for item in best_items:
                history_items_recommended.append((uid, item))
                model.observe(
                    None, (uid, item), test_consumption_matrix[uid, item], info
                )
            model.increment_time()
            users_num_interactions[uid] += 1
            if users_num_interactions[uid] == self.num_interactions:
                available_users = available_users - {uid}

            _num_interactions += 1
            if i % _intervals == 0 and i != 0:
                pbar.update(_num_interactions)
                _num_interactions = 0

        pbar.update(_num_interactions)
        _num_interactions = 0
        pbar.close()
        items_entropy = irec.value_functions.Entropy.get_items_entropy(
            consumption_matrix
        )
        items_popularity = irec.value_functions.MostPopular.get_items_popularity(
            consumption_matrix, normalize=False
        )
        for uid, items in users_items_recommended.items():

            colors = mpl.cm.rainbow(np.linspace(0, 1, len(items)))
            fig = plt.figure(figsize=(8, 5))
            # plt.colorbar(colors)
            plt.rcParams.update({"font.size": 14})
            plt.scatter(items_entropy, items_popularity, s=100, color="#d1d1d1")

            plt.scatter(
                items_entropy[items], items_popularity[items], s=100, color=colors
            )

            plt.text(
                0,
                1.5,
                "Correlation: %.4f"
                % scipy.stats.pearsonr(items_entropy, items_popularity)[0],
                bbox={"facecolor": "red", "alpha": 0.5, "pad": 5},
            )
            plt.xlabel("Entropy")
            plt.ylabel("Popularity")
            fig.savefig(
                os.path.join(DirectoryDependent().DIRS["img"], f"plot_{uid}.png"),
                bbox_inches="tight",
            )

        return history_items_recommended


class LimitedInteraction(EvaluationPolicy):
    def __init__(
        self, interaction_size, recommend_test_data_rate_limit, *args, **kwargs
    ):
        super().__init__(*args, **kwargs)
        self.interaction_size = interaction_size
        self.recommend_test_data_rate_limit = recommend_test_data_rate_limit

    def evaluate(self, model, train_dataset, test_dataset):
        with threadpool_limits(limits=1, user_api="blas"):
            test_users = np.unique(test_dataset.data[:, 0]).astype(int)
            num_total_items = test_dataset.num_total_items
            test_consumption_matrix = scipy.sparse.csr_matrix(
                (
                    test_dataset.data[:, 2],
                    (
                        test_dataset.data[:, 0].astype(int),
                        test_dataset.data[:, 1].astype(int),
                    ),
                ),
                shape=(test_dataset.num_total_users, test_dataset.num_total_items),
            )

            users_items_recommended = defaultdict(list)
            num_test_users = len(test_users)
            print(f"Starting {model.name} Training")
            model.reset(train_dataset)
            print(f"Ended {model.name} Training")
            # users_num_interactions = defaultdict(int)
            users_num_items_to_recommend_from_test = dict()
            available_users = set()
            for uid in test_users:
                users_num_items_to_recommend_from_test[uid] = np.floor(
                    (test_consumption_matrix[uid] > 0).count_nonzero()
                    * self.recommend_test_data_rate_limit
                )
                if users_num_items_to_recommend_from_test[uid] > 0:
                    available_users |= {uid}

            users_num_items_recommended_from_test = defaultdict(int)

            history_items_recommended = []

            while len(available_users) > 0:
                uid = random.sample(available_users, k=1)[0]
                not_recommended = np.ones(num_total_items, dtype=bool)
                not_recommended[users_items_recommended[uid]] = 0
                items_not_recommended = np.nonzero(not_recommended)[0]
                # items_score, info = model.action_estimates(
                # (uid, items_not_recommended))
                # best_items = items_not_recommended[np.argpartition(
                # items_score,
                # -self.interaction_size)[-self.interaction_size:]]
                actions, info = model.act(
                    (uid, items_not_recommended), self.interaction_size
                )
                best_items = actions[1]
                users_items_recommended[uid].extend(best_items)

                for item in best_items:
                    history_items_recommended.append((uid, item))
                    model.observe(
                        None, (uid, item), test_consumption_matrix[uid, item], info
                    )
                    users_num_items_recommended_from_test[uid] += (
                        test_consumption_matrix[uid, item] > 0
                    )

                # users_num_interactions[uid] += 1
                if (
                    users_num_items_recommended_from_test[uid]
                    >= users_num_items_to_recommend_from_test[uid]
                ):
                    available_users = available_users - {uid}

            return history_items_recommended, None


class OneInteraction(EvaluationPolicy):
    def __init__(self, num_interactions, interaction_size, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.num_interactions = num_interactions
        self.interaction_size = interaction_size

    def evaluate(self, model, train_dataset, test_dataset):
        with threadpool_limits(limits=1, user_api="blas"):
            test_users = np.unique(test_dataset.data[:, 0]).astype(int)
            num_total_items = test_dataset.num_total_items
            test_consumption_matrix = scipy.sparse.csr_matrix(
                (
                    test_dataset.data[:, 2],
                    (
                        test_dataset.data[:, 0].astype(int),
                        test_dataset.data[:, 1].astype(int),
                    ),
                ),
                shape=(test_dataset.num_total_users, test_dataset.num_total_items),
            )

            users_items_recommended = defaultdict(list)

            for i in range(len(train_dataset.data)):
                uid = int(train_dataset.data[i, 0])
                if uid in test_users:
                    iid = int(train_dataset.data[i, 1])
                    reward = train_dataset.data[i, 2]
                    users_items_recommended[uid].append(iid)

            num_test_users = len(test_users)
            print(f"Starting {model.name} Training")
            model.reset(train_dataset)
            print(f"Ended {model.name} Training")
            users_num_interactions = defaultdict(int)
            available_users = set(test_users)

            history_items_recommended = []

            num_trials = num_test_users * self.num_interactions
            _intervals = num_trials // 20
            _num_interactions = 0
            pbar = tqdm(total=num_trials)
            pbar.set_description(f"{model.name}")
            no_items_recommended_users = set(test_users)

            train_consumption_matrix = scipy.sparse.csr_matrix(
                (
                    train_dataset.data[:, 2],
                    (train_dataset.data[:, 0], train_dataset.data[:, 1]),
                ),
                (train_dataset.num_total_users, train_dataset.num_total_items),
            )

            items_entropy = irec.value_functions.Entropy.get_items_entropy(
                train_consumption_matrix
            )
            items_popularity = irec.value_functions.MostPopular.get_items_popularity(
                train_consumption_matrix, normalize=True
            )
            del train_consumption_matrix
            # correlations = defaultdict(dict)
            items_value = defaultdict(dict)
            items_value_table = []
            membership = defaultdict(dict)
            # membership_table = [['Dataset','Method','Pop. Corr.','Ent. Corr.']]
            membership_table = []
            top_k_nonp = 100
            top_k_nonp_items_popularity = np.argsort(items_popularity)[::-1][
                :top_k_nonp
            ]
            top_k_nonp_items_entropy = np.argsort(items_entropy)[::-1][:top_k_nonp]

            for i in range(num_trials):
                uid = random.sample(available_users, k=1)[0]
                not_recommended = np.ones(num_total_items, dtype=bool)
                not_recommended[users_items_recommended[uid]] = 0
                items_not_recommended = np.nonzero(not_recommended)[0]
                items_score, info = model.action_estimates((uid, items_not_recommended))
                best_items = items_not_recommended[
                    np.argpartition(items_score, -self.interaction_size)[
                        -self.interaction_size :
                    ]
                ]
                users_items_recommended[uid].extend(best_items)

                for item in best_items:
                    history_items_recommended.append((uid, item))
                    model.observe(
                        None, (uid, item), test_consumption_matrix[uid, item], info
                    )
                model.increment_time()

                users_num_interactions[uid] += 1
                if users_num_interactions[uid] == self.num_interactions:
                    available_users = available_users - {uid}

                if users_num_interactions[uid] == 1:
                    # correlations['Popularity'][uid]= scipy.stats.pearsonr(items_score,items_popularity)[0]
                    # correlations['Entropy'][uid]= scipy.stats.pearsonr(items_score,items_entropy)[0]
                    items_value["Popularity"][uid] = np.mean(
                        items_popularity[best_items]
                    )
                    items_value["Entropy"][uid] = np.mean(items_entropy[best_items])
                    membership["Popularity"][uid] = len(
                        set(best_items).intersection(set(top_k_nonp_items_popularity))
                    )
                    membership["Entropy"][uid] = len(
                        set(best_items).intersection(set(top_k_nonp_items_entropy))
                    )
                    no_items_recommended_users -= {uid}

                _num_interactions += 1
                if i % _intervals == 0 and i != 0:
                    pbar.update(_num_interactions)
                    _num_interactions = 0

                if len(no_items_recommended_users) == 0:
                    print(f"Finished recommendations in {i+1} trial")
                    base_name = "None"
                    if train_dataset.num_total_users == 69878:
                        base_name = "MovieLens 10M"
                    elif train_dataset.num_total_users == 53423:
                        base_name = "Good Books"
                    elif train_dataset.num_total_users == 15400:
                        base_name = "Yahoo Music"
                    # items_value_table.append(base_name,model.name,*[np.mean(list(corr_values.values())) for corr_name, corr_values in items_value.items()])
                    # membership_table.append(base_name,model.name,*[np.mean(list(corr_values.values())) for corr_name, corr_values in membership.items()])
                    # print('Correlation: {} {} {} {}'.format(train_dataset.num_total_users,train_dataset.num_total_items,model.name,
                    # ' '.join(['{} {}'.format(corr_name,np.mean(list(corr_values.values()))) for corr_name, corr_values in correlations.items() ])))
                    # print('Items_Values: {} {} {} {}'.format(train_dataset.num_total_users,train_dataset.num_total_items,model.name,
                    # ' '.join(['{} {}'.format(corr_name,np.mean(list(corr_values.values()))) for corr_name, corr_values in items_value.items() ])
                    # ))
                    print(
                        "Items_Values: {} {} {} {}".format(
                            train_dataset.num_total_users,
                            train_dataset.num_total_items,
                            model.name,
                            " ".join(
                                [
                                    "{} {}".format(
                                        corr_name, np.mean(list(corr_values.values()))
                                    )
                                    for corr_name, corr_values in items_value.items()
                                ]
                            ),
                        )
                    )
                    print(
                        "Membership: {} {} {} {}".format(
                            train_dataset.num_total_users,
                            train_dataset.num_total_items,
                            model.name,
                            " ".join(
                                [
                                    "{} {}".format(
                                        corr_name, np.mean(list(corr_values.values()))
                                    )
                                    for corr_name, corr_values in membership.items()
                                ]
                            ),
                        )
                    )
                    break

            # print(tabulate(items_value_table, tablefmt='psql'))
            # print(tabulate(membership_table, tablefmt='psql'))
            pbar.update(_num_interactions)
            _num_interactions = 0
            pbar.close()
            return history_items_recommended


class PercentageInteraction(EvaluationPolicy):
    def __init__(
        self,
        num_interactions: int,
        interaction_size: int,
        *args,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        self.num_interactions = int(num_interactions)
        self.interaction_size = int(interaction_size)

    @staticmethod
    def run_eval(parameters):
        obj_id, uid = parameters[0], parameters[1]
        self = ctypes.cast(obj_id, ctypes.py_object).value

        history_items_recommended = {}
        for method in self.nonp_methods:
            history_items_recommended[method] = {}
            
            for exchange_point in self.exchange_points:
                history_items_recommended[method][exchange_point] = []

                t, num_items = 0, 0
                max_items = int(len(self.test_consumption_matrix[uid,:].data) * exchange_point)
                user_items_recommended = self.users_items_recommended_original[uid].copy()
                not_recommended = np.ones(self.num_total_items, dtype=bool)
                not_recommended[user_items_recommended] = 0
                items_not_recommended = np.nonzero(not_recommended)[0]

                #rec nonpers
                top_k_items = random.sample(list(items_not_recommended), self.num_total_items) if method == "top_k_items_random" else self.nonp_methods[method]
                for item in top_k_items:

                    if self.test_consumption_matrix[uid, item] != 0:
                        self.model.observe(None, (uid, item), self.test_consumption_matrix[uid, item], {"vf_info": None, "asp_info": None})
                        num_items += 1

                    user_items_recommended.append(item)
                    not_recommended[item] = 0
                    if num_items > max_items: break
                        
                del top_k_items
                items_not_recommended = np.nonzero(not_recommended)[0]

                #rec mab
                for itr in range(self.num_interactions):
                    actions, info = self.model.act(OneUserCandidateActions(uid, items_not_recommended), self.interaction_size)
                    best_items = actions[1]
                    for item in best_items:
                        not_recommended[item] = 0
                        user_items_recommended.append(item)
                        if self.test_consumption_matrix[uid, item] != 0: 
                            self.model.observe(None, (uid, item), self.test_consumption_matrix[uid, item], info)

                    items_not_recommended = np.nonzero(not_recommended)[0]

                history_items_recommended[method][exchange_point].append((uid, user_items_recommended))
                del not_recommended, items_not_recommended
        return history_items_recommended

    def evaluate(self, model, train_dataset, test_dataset):
        
        # with threadpool_limits(limits=1, user_api="blas"):
        test_users = np.unique(test_dataset.data[:, 0]).astype(int)
        self.model = model
        self.num_total_items = test_dataset.num_total_items
        self.test_consumption_matrix = scipy.sparse.csr_matrix(
            (
                test_dataset.data[:, 2],
                (
                    test_dataset.data[:, 0].astype(int),
                    test_dataset.data[:, 1].astype(int),
                ),
            ),
            shape=(test_dataset.num_total_users, test_dataset.num_total_items),
        )

        self.users_items_recommended_original = defaultdict(list)

        for i in range(len(train_dataset.data)):
            uid = int(train_dataset.data[i, 0])
            if uid in test_users:
                iid = int(train_dataset.data[i, 1])
                reward = train_dataset.data[i, 2]
                self.users_items_recommended_original[uid].append(iid)

        num_test_users = len(test_users)
        print(f"Starting {self.model.name} Training")
        self.model.reset(train_dataset)
        print(f"Ended {model.name} Training")
        available_users = set(test_users)

        train_consumption_matrix = scipy.sparse.csr_matrix(
            (
                train_dataset.data[:, 2],
                (train_dataset.data[:, 0], train_dataset.data[:, 1]),
            ),
            (train_dataset.num_total_users, train_dataset.num_total_items),
        )

        items_entropy = Entropy.get_items_entropy(train_consumption_matrix)
        items_popularity = MostPopular.get_items_popularity(train_consumption_matrix, normalize=False)
        items_logPopEnt = LogPopEnt.get_items_logpopent(items_popularity, items_entropy)
        items_bestRated = BestRated.get_items_bestrated(train_consumption_matrix)

        self.exchange_points = [0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8]
        self.nonp_methods = {
            "top_k_items_popularity": np.argsort(items_popularity)[::-1].astype(np.int32),
            "top_k_items_entropy": np.argsort(items_entropy)[::-1].astype(np.int32),
            "top_k_items_logPopEnt": np.argsort(items_logPopEnt)[::-1].astype(np.int32),
            "top_k_items_bestRated": np.argsort(items_bestRated)[::-1].astype(np.int32),
            "top_k_items_random": None
        }

        self_id = id(self)
        parameters = [[self_id], available_users]
        parameters = list(itertools.product(*parameters)) 

        executor = ProcessPoolExecutor()
        num_args = len(parameters)
        chunksize = int(num_args/multiprocessing.cpu_count())

        history_items_recommended = [i for i in tqdm(executor.map(PercentageInteraction.run_eval, parameters),total=num_args)]
        
        return history_items_recommended, None



class RelevanceEvaluator:
    """RelevanceEvaluator."""

    def __init__(self, *args, **kwargs):
        """__init__.

        Args:
            args:
            kwargs:
        """
        super().__init__(*args, **kwargs)

    def is_relevant(self, reward: float):
        """is_relevant.

        Args:
            reward (float): reward
        """
        raise NotImplemented


class ThresholdRelevanceEvaluator:
    """ThresholdRelevanceEvaluator."""

    def __init__(self, threshold: float, *args, **kwargs):
        """__init__.

        Args:
            threshold (float): threshold
            args:
            kwargs:
        """
        super().__init__(*args, **kwargs)
        self.threshold = threshold

    def is_relevant(self, reward: float):
        """is_relevant.

        Args:
            reward (float): reward
        """
        return reward > self.threshold



from .ExperimentalValueFunction import ExperimentalValueFunction
import numpy as np
from tqdm import tqdm
#import util
from threadpoolctl import threadpool_limits
import scipy
import ctypes
import mf
from collections import defaultdict
from .MFValueFunction import MFValueFunction


class LinUCBvar(MFValueFunction):
    def __init__(self, alpha, lambda_u, zeta=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if alpha != None:
            self.alpha = alpha
        elif zeta != None:
            self.alpha = 1 + np.sqrt(np.log(2 / zeta) / 2)
        self.lambda_u = lambda_u


    def reset(self, observation):
        train_dataset = observation
        super().reset(train_dataset)
        self.train_dataset = train_dataset
        self.train_consumption_matrix = scipy.sparse.csr_matrix(
            (self.train_dataset.data[:, 2],
             (self.train_dataset.data[:, 0], self.train_dataset.data[:, 1])),
            (self.train_dataset.num_total_users,
             self.train_dataset.num_total_items))
        self.num_total_items = self.train_dataset.num_total_items

        mf_model = mf.SVD(num_lat=self.num_lat)
        mf_model.fit(self.train_consumption_matrix)
        self.items_weights = mf_model.items_weights
        self.num_latent_factors = len(self.items_weights[0])

        self.I = np.eye(len(self.items_weights[0]))
        self.bs = defaultdict(lambda: np.ones(self.num_latent_factors))
        self.As = defaultdict(lambda: self.lambda_u * self.I.copy())

    def action_estimates(self, candidate_actions):
        uid = candidate_actions[0]
        candidate_items = candidate_actions[1]
        b = self.bs[uid]
        A = self.As[uid]
        mean = np.dot(np.linalg.inv(A), b)
        items_uncertainty = self.alpha * np.sqrt(
            np.sum(self.items_weights[candidate_items].dot(np.linalg.inv(A)) *
                   self.items_weights[candidate_items],
                   axis=1))
        items_user_similarity = mean @ self.items_weights[candidate_items].T
        items_score = items_user_similarity + items_uncertainty

        return items_score, None

    def update(self, observation, action, reward, info):
        uid = action[0]
        item = action[1]
        additional_data = info
        max_item_latent_factors = self.items_weights[item]
        b = self.bs[uid]
        A = self.As[uid]
        A += max_item_latent_factors[:, None].dot(
            max_item_latent_factors[None, :])
        b += reward * max_item_latent_factors



class CandidateAction:
    """CandidateAction."""

    pass


class UICandidateAction(CandidateAction):
    """UICandidateAction."""

    def __init__(self, user: int, item: int):
        """__init__.

        Args:
            user (int): user
            item (int): item
        """
        self.user = user
        self.item = item



from typing import List


class CandidateActions:
    """CandidateActions."""

    pass


class OneUserCandidateActions(CandidateActions):
    """OneUserCandidateActions."""

    def __init__(self, user: int, items: List[int]) -> None:
        """__init__.

        Args:
            user (int): user
            items (List[int]): items

        Returns:
            None:
        """
        super().__init__()
        self.user = user
        self.items = items

    def __getitem__(self, key):
        if key == 0:
            return self.user
        elif key == 1:
            return self.items
        else:
            raise IndexError



import numpy as np
from tqdm import tqdm

# import util
from threadpoolctl import threadpool_limits
import ctypes
import scipy.spatial
import matplotlib.pyplot as plt
import os
import pickle
import sklearn
import scipy.optimize
import scipy
import scipy.stats
import mf
from collections import defaultdict
from .MFValueFunction import MFValueFunction
import value_functions
from value_functions.MostPopular import MostPopular
from value_functions.Entropy import Entropy
from value_functions.LogPopEnt import LogPopEnt


def _prediction_rule(A, b, items_weights, alpha):
    user_latent_factors = np.dot(np.linalg.inv(A), b)
    items_uncertainty = np.sqrt(
        np.sum(items_weights.dot(np.linalg.inv(A)) * items_weights, axis=1)
    )
    items_user_similarity = user_latent_factors @ items_weights.T
    weighted_items_uncertainty = alpha * items_uncertainty
    user_model_items_score = items_user_similarity + weighted_items_uncertainty
    return user_model_items_score, items_user_similarity, weighted_items_uncertainty


class OurMethodInit(MFValueFunction):
    def __init__(self, alpha, init, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.alpha = alpha
        self.init = init

    def reset(self, observation):
        train_dataset = observation
        super().reset(train_dataset)
        self.train_dataset = train_dataset
        self.train_consumption_matrix = scipy.sparse.csr_matrix(
            (
                self.train_dataset.data[:, 2],
                (self.train_dataset.data[:, 0], self.train_dataset.data[:, 1]),
            ),
            (self.train_dataset.num_total_users, self.train_dataset.num_total_items),
        )
        self.num_total_items = self.train_dataset.num_total_items

        mf_model = mf.SVD(num_lat=self.num_lat)
        mf_model.fit(self.train_consumption_matrix)
        self.items_weights = mf_model.items_weights
        self.num_latent_factors = len(self.items_weights[0])
        if self.init == "zero":
            self.bs = defaultdict(lambda: np.zeros(self.num_lat))
        elif self.init == "one":
            self.bs = defaultdict(lambda: np.ones(self.num_lat))

        if self.init in [
            "entropy",
            "popularity",
            "logpopent",
            "rand_popularity",
            "random",
        ]:
            if self.init == "entropy":
                items_entropy = Entropy.get_items_entropy(self.train_consumption_matrix)
                self.items_bias = items_entropy
            elif self.init == "popularity":
                items_popularity = MostPopular.get_items_popularity(
                    self.train_consumption_matrix, normalize=False
                )
                self.items_bias = items_popularity
            elif self.init == "logpopent":
                items_entropy = Entropy.get_items_entropy(self.train_consumption_matrix)
                items_popularity = MostPopular.get_items_popularity(
                    self.train_consumption_matrix, normalize=False
                )
                self.items_bias = LogPopEnt.get_items_logpopent(
                    items_popularity, items_entropy
                )
            elif self.init == "rand_popularity":
                items_popularity = MostPopular.get_items_popularity(
                    self.train_consumption_matrix, normalize=False
                )
                items_popularity[np.argsort(items_popularity)[::-1][100:]] = 0
                self.items_bias = items_popularity
            elif self.init == "random":
                self.items_bias = np.random.rand(self.train_dataset.num_total_items)

            self.items_bias = self.items_bias - np.min(self.items_bias)
            self.items_bias = self.items_bias / np.max(self.items_bias)

            assert self.items_bias.min() >= 0 and np.isclose(self.items_bias.max(), 1)

            # regression_model = sklearn.linear_model.LinearRegression()
            res = scipy.optimize.minimize(
                lambda x, items_weights, items_bias: np.sum(
                    (items_bias - x @ items_weights.T) ** 2
                ),
                np.ones(self.num_latent_factors),
                args=(self.items_weights, self.items_bias),
                method="BFGS",
            )
            self.initial_b = res.x

            print(
                np.corrcoef(self.items_bias, self.initial_b @ self.items_weights.T)[
                    0, 1
                ]
            )

            self.bs = defaultdict(lambda: self.initial_b.copy())
        self.I = np.eye(len(self.items_weights[0]))
        self.As = defaultdict(lambda: self.I.copy())

        # self.users_last_items = dict()
        # items_score = _prediction_rule(self.I,self.initial_b,self.items_weights,self.alpha)
        # items_entropy = value_functions.Entropy.get_items_entropy(self.train_consumption_matrix)
        # items_popularity = value_functions.MostPopular.get_items_popularity(self.train_consumption_matrix,normalize=False)

        # print(f"WSCB {self.init} items score correlation with popularity:",scipy.stats.pearsonr(items_score,items_popularity),self.train_dataset.num_total_users, self.train_dataset.num_total_items)
        # print(f"WSCB {self.init} items score correlation with entropy:",scipy.stats.pearsonr(items_score,items_entropy),self.train_dataset.num_total_users, self.train_dataset.num_total_items)

    def action_estimates(self, candidate_actions):
        uid = candidate_actions[0]
        candidate_items = candidate_actions[1]
        b = self.bs[uid]
        A = self.As[uid]
        user_model_items_score, l1, l2 = _prediction_rule(
            A, b, self.items_weights[candidate_items], self.alpha
        )
        items_score = user_model_items_score
        return items_score, {
            "class_name": self.__class__.__name__,
            "uid": uid,
            "items_score":items_score,
            "similarity": scipy.stats.describe(l1),
            "uncertainty": scipy.stats.describe(l2),
        }

    def update(self, observation, action, reward, info):
        uid = action[0]
        item = action[1]
        additional_data = info
        max_item_latent_factors = self.items_weights[item]
        b = self.bs[uid]
        A = self.As[uid]
        A += max_item_latent_factors[:, None].dot(max_item_latent_factors[None, :])
        b += reward * max_item_latent_factors


class OurMethodEntropy(OurMethodInit):
    def __init__(self, *args, **kwargs):
        super().__init__(init="entropy", *args, **kwargs)


class OurMethodPopularity(OurMethodInit):
    def __init__(self, *args, **kwargs):
        super().__init__(init="popularity", *args, **kwargs)


class OurMethodRandPopularity(OurMethodInit):
    def __init__(self, *args, **kwargs):
        super().__init__(init="rand_popularity", *args, **kwargs)


class OurMethodRandom(OurMethodInit):
    def __init__(self, *args, **kwargs):
        super().__init__(init="random", *args, **kwargs)


class OurMethodZero(OurMethodInit):
    def __init__(self, *args, **kwargs):
        super().__init__(init="zero", *args, **kwargs)


class OurMethodOne(OurMethodInit):
    def __init__(self, *args, **kwargs):
        super().__init__(init="one", *args, **kwargs)



from concurrent.futures import ProcessPoolExecutor
import math
import collections
from tqdm import tqdm
import os
import multiprocessing
import numpy as np
from pathlib import Path


class TupleNonRedundantList():
    def __init__(self, redundant_key, non_redundant_list) -> None:
        self.non_redundant_list = non_redundant_list
        self.redundant_key = redundant_key
        pass

    def __delitem__(self, key):
        # self.non_redundant_list[1].__delattr__(key)
        raise NotImplementedError

    def __getitem__(self, key):
        return (self.redundant_key,
                self.non_redundant_list.__getattribute__(key))

    def __setitem__(self, key, value):
        raise NotImplementedError
        # self.non_redundant_list[1].__setattr__(key, value)


def dict_to_list_gen(d):
    for k, v in zip(d.keys(), d.values()):
        if v == None:
            continue
        yield k
        yield v


def dict_to_list(d):
    return list(dict_to_list_gen(d))


def value_to_str(value):
    if isinstance(value, dict):
        return f"{{{dict_to_str(value)}}}"
    elif isinstance(value, list):
        return f"{join_strings(list(map(lambda x: value_to_str(x), value)))}"
    else:
        return f"{str(value).replace('/','|')}"


def key_value_to_str(key, value):
    return f"{key}:{value_to_str(value)}"


def join_strings(strings, num_bars=0):
    return "/".join(
        strings[:num_bars]) + ("/" if num_bars and len(strings[num_bars:]) != 0
                               else "") + ",".join(strings[num_bars:])


def dict_to_str(dictionary, num_bars=0):
    strings = []
    for key, value in dictionary.items():
        strings.append(key_value_to_str(key, value))

    return join_strings(strings, num_bars=num_bars)


def print_dict(dictionary, prefix=''):
    for key, value in dictionary.items():
        if isinstance(value, dict):
            print(f"{prefix}{key}:")
            print_dict(value, prefix + '\t')
        else:
            print(f"{prefix}{key}: {value}")


def run_parallel(func, args, use_tqdm=True):
    executor = ProcessPoolExecutor()
    num_args = len(args)
    chunksize = math.ceil(num_args / multiprocessing.cpu_count())
    if use_tqdm:
        ff = tqdm
    else:
        ff = lambda x, *y, **z: x
    results = [
        i
        for i in ff(executor.map(func, *list(zip(*args)), chunksize=chunksize),
                    total=num_args)
    ]
    return results


def sigmoid(x):
    return 1 / (1 + np.exp(-x))



def repair_path_name(path):
    length = len(path)
    new_file_name = []
    for i in path.split('/'):
        if len(i) > 255:
            lists = [i[j:j + 255] for j in range(0, len(i), 255)]
            new_file_name.append('/'.join(lists))
        else:
            new_file_name.append(i)
    return '/'.join(new_file_name)



import numpy as np
from tqdm import tqdm
from .ExperimentalValueFunction import ExperimentalValueFunction
import matplotlib.pyplot as plt
import os
import scipy
from collections import defaultdict
import random
import itertools
from numba import jit, prange
import mf
from .MFValueFunction import MFValueFunction


class COFIBA(MFValueFunction):
    """COFIBA.
    
    This method relies on upper-confidence-based tradeoffs between exploration and exploitation,
    combined with adaptive clustering procedures at both the user and the item sides [1]_.

    References
    ----------
    .. [1] Li, Shuai, Alexandros Karatzoglou, and Claudio Gentile. "Collaborative filtering bandits." 
       Proceedings of the 39th International ACM SIGIR conference on Research and Development 
       in Information Retrieval. 2016.   
    """
    def __init__(self, alpha=1, alpha_2=1, *args, **kwargs):
        """__init__.

        Args:
            args:
            kwargs:
            alpha:
            alpha_2:
        """
        super().__init__(*args, **kwargs)
        self.alpha = alpha
        self.alpha_2 = alpha_2


    def cb(self, alpha, item_latent_factors, m, t):
        """cb.

        Args:
            alpha: 
            item_latent_factors:
            m: 
            t: 

        Returns:
            float
        """
        return alpha * np.sqrt(item_latent_factors.T @ np.linalg.inv(m)
                               @ item_latent_factors * np.log10(t + 1))
        pass

    def update_user_cluster(self, uid, item):
        """update_user_cluster.

        Args:
            uid (int): user id
            item (int): item id

        Returns:
            :
        """
        item_cluster = self.items_clustering[item]
        users_graph = self.users_graphs[item_cluster].copy()
        neighbors = np.nonzero(users_graph[uid])[1]
        for neighbor in neighbors:
            if np.abs(self.users_latent_factors[uid] @ self.items_latent_factors[item] - self.users_latent_factors[neighbor] @ self.items_latent_factors[item])\
               > self.cb(self.alpha_2,self.items_latent_factors[item],self.users_m[uid], self.t) + self.cb(self.alpha_2,self.items_latent_factors[item],self.users_m[neighbor], self.t):
                users_graph[uid, neighbor] = 0
                users_graph[neighbor, uid] = 0
        n_components, labels = scipy.sparse.csgraph.connected_components(
            users_graph)
        # self.users_clusterings[item_cluster] = labels
        return users_graph, labels

    def update_item_cluster(self, uid, item):
        """update_item_cluster.

        Args:
            uid (int): user id
            item (int): item id

        """
        item_cluster = self.items_clustering[item]
        actual_cluster_items = set(
            np.nonzero(self.items_clustering == item_cluster)[0])

        neighbors = np.nonzero(self.items_graph[item])[1]
        # users_graph = self.users_graphs[item_cluster].copy()
        # neighbors = np.nonzero(users_graph[uid])[1]

        generated_user_neighbors = []
        for neighbor in neighbors:
            generated_user_neighbors = set()
            for uid2 in range(self.num_total_users):
                if np.abs(self.users_latent_factors[uid] @ self.items_latent_factors[neighbor] - self.users_latent_factors[uid2] @ self.items_latent_factors[neighbor])\
                <= self.cb(self.alpha_2,self.items_latent_factors[neighbor],self.users_m[uid], self.t) + self.cb(self.alpha_2,self.items_latent_factors[neighbor],self.users_m[uid2], self.t):
                    generated_user_neighbors.add(uid)
            if generated_user_neighbors != actual_cluster_items:
                self.items_graph[item, neighbor] = 0
                self.items_graph[neighbor, item] = 0

        n_components, labels = scipy.sparse.csgraph.connected_components(
            self.items_graph)
        self.items_clustering = labels
        r = range(self.items_n_components, n_components)
        self.items_n_components = n_components
        for i in r:
            users_graph = self.new_graph(self.num_total_users)
            self.users_graphs.append(users_graph)
            n_components, labels = scipy.sparse.csgraph.connected_components(
                users_graph)
            self.users_clusterings.append(labels)

    @staticmethod
    def new_graph(n):
        """new_graph.

        Args:
            n (int):

        Returns:
            sparse matrix:
        """
        graph = scipy.sparse.random(n,
                                    n,
                                    density=2 * np.log(n) / n,
                                    dtype=bool,
                                    format='lil')
        COFIBA.symmetrize_matrix(graph)
        for i in range(graph.shape[0]):
            graph[i, i] = 0
        graph.tocsr().eliminate_zeros()
        return graph

    @staticmethod
    def symmetrize_matrix(m):
        """symmetrize_matrix.

        Args:
            m (int):

        """
        for i in range(m.shape[0]):
            for j in range(m.shape[0]):
                if j < i:
                    m[j, i] = m[i, j]

    def score(self, uid, item, user_connected_component):
        """score.

        Args:
            uid (int): user id
            item (int): item id
            user_connected_component:

        Returns:
            :
        """
        neighbors = user_connected_component
        num_neighbors = len(neighbors)
        cluster_m = self.I + np.add.reduce(self.users_m[np.append(
            neighbors, uid)]) - num_neighbors * self.I
        cluster_b = np.add.reduce(self.users_b[np.append(neighbors, uid)])
        cluster_latent_factors = cluster_m @ cluster_b
        return cluster_latent_factors @ self.items_latent_factors[
            item] + self.cb(self.alpha, self.items_latent_factors[item],
                            cluster_m, self.t)

    def reset(self, observation):
        """reset.

        Args:
            observation: 
        """
        train_dataset = observation
        super().reset(train_dataset)
        self.train_dataset = train_dataset
        self.train_consumption_matrix = scipy.sparse.csr_matrix(
            (self.train_dataset.data[:, 2],
             (self.train_dataset.data[:, 0], self.train_dataset.data[:, 1])),
            (self.train_dataset.num_total_users,
             self.train_dataset.num_total_items))
        self.num_total_items = self.train_dataset.num_total_items
        self.consumption_matrix = self.train_consumption_matrix.tolil()
        self.num_total_users = self.train_dataset.num_total_users

        mf_model = mf.SVD(num_lat=self.num_lat)
        mf_model.fit(self.train_consumption_matrix)
        self.items_latent_factors = mf_model.items_weights

        self.I = np.identity(self.num_lat)

        self.items_graph = self.new_graph(self.num_total_items)

        self.items_n_components, self.items_clustering = scipy.sparse.csgraph.connected_components(
            self.items_graph)
        self.users_graphs = []
        self.users_clusterings = []
        for i in range(self.items_n_components):
            users_graph = self.new_graph(self.num_total_users)
            self.users_graphs.append(users_graph)
            n_components, labels = scipy.sparse.csgraph.connected_components(
                users_graph)
            self.users_clusterings.append(labels)
        self.users_b = np.zeros((self.num_total_users, self.num_lat))
        self.users_m = []
        for i in range(self.num_total_users):
            self.users_m.append(np.identity(self.num_lat))

        self.users_m = np.array(self.users_m)
        self.users_latent_factors = [
            np.linalg.inv(m) @ b for b, m in zip(self.users_b, self.users_m)
        ]
        self.t = 1
        self.recent_predict = True

    def action_estimates(self, candidate_actions):
        """action_estimates.

        Args:
            candidate_actions: (user id, candidate_items)

        Returns:
            numpy.ndarray:
        """
        uid = candidate_actions[0]
        candidate_items = candidate_actions[1]
        items_score = np.zeros(candidate_items.shape)
        for i, item in enumerate(candidate_items):
            users_graph, labels = self.update_user_cluster(uid, item)
            user_connected_component = np.nonzero(labels[uid] == labels)[0]
            items_score[i] = self.score(uid, item, user_connected_component)
        self.recent_predict = True
        return items_score, None

    def update(self, observation, action, reward, info):
        """update.

        Args:
            observation:
            action: (user id, item)
            reward (float): reward
            info: 
        """
        uid = action[0]
        item = action[1]
        additional_data = info
        users_graph, labels = self.update_user_cluster(uid, item)
        item_cluster = self.items_clustering[item]
        self.users_graphs[item_cluster] = users_graph
        self.update_item_cluster(uid, item)
        if self.recent_predict:
            self.t += 1
            self.recent_predict = False



import numpy as np
from tqdm import tqdm
from .ExperimentalValueFunction import ExperimentalValueFunction
import matplotlib.pyplot as plt
import os
import scipy.sparse


class Entropy(ExperimentalValueFunction):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    @staticmethod
    def probabilities_entropy(probabilities):
        try:
            return -1 * np.sum(probabilities * np.log(probabilities))
        except:
            return 0

    @staticmethod
    def values_entropy(values):
        unique, counts = np.unique(values, return_counts=True)
        values_probability = counts / np.sum(counts)
        return Entropy.probabilities_entropy(values_probability)

    @staticmethod
    def get_items_entropy(consumption_matrix):
        lowest_value = np.min(consumption_matrix)
        items_entropy = np.zeros(consumption_matrix.shape[1])
        is_spmatrix = isinstance(consumption_matrix, scipy.sparse.spmatrix)
        if is_spmatrix:
            consumption_matrix = scipy.sparse.csc_matrix(consumption_matrix)
        for iid in range(consumption_matrix.shape[1]):
            if is_spmatrix:
                iid_ratings = consumption_matrix[:, iid].A.flatten()
                iid_ratings = iid_ratings[iid_ratings > lowest_value]
            else:
                iid_ratings = consumption_matrix[:, iid]
                iid_ratings = iid_ratings[iid_ratings > lowest_value]
            items_entropy[iid] = Entropy.values_entropy(iid_ratings)
        return items_entropy

    def reset(self, observation):
        train_dataset = observation
        super().reset(train_dataset)
        self.train_dataset = train_dataset
        self.num_total_items = self.train_dataset.num_total_items
        self.unique_values = self.train_dataset.rate_domain
        self.num_unique_values = len(self.unique_values)
        self.items_ratings = np.zeros(
            (self.num_total_items, self.num_unique_values))
        self.unique_values_ids = dict(
            zip(self.unique_values, list(range(self.num_unique_values))))
        self.items_num_total_ratings = np.zeros(self.num_total_items)
        for uid, iid, reward, *rest in self.train_dataset.data:
            self.items_ratings[int(iid), self.unique_values_ids[reward]] += 1
            self.items_num_total_ratings[int(iid)] += 1

    def action_estimates(self, candidate_actions):
        uid = candidate_actions[0]
        candidate_items = candidate_actions[1]
        items_score = [
            self.probabilities_entropy(self.items_ratings[iid] /
                                       np.sum(self.items_ratings[iid]))
            if self.items_num_total_ratings[iid] > 0 else 0
            for iid in candidate_items
        ]
        return items_score, None

    def update(self, observation, action, reward, info):
        uid = action[0]
        item = action[1]
        additional_data = info
        if reward in self.unique_values_ids:
            self.items_ratings[item, self.unique_values_ids[reward]] += 1
            self.items_num_total_ratings[item] += 1



from .ExperimentalValueFunction import ExperimentalValueFunction
import numpy as np
import random
from tqdm import tqdm
#import util
from threadpoolctl import threadpool_limits
import ctypes
from .Entropy import Entropy
from .MostPopular import MostPopular
from .LogPopEnt import LogPopEnt
from .PopPlusEnt import *
from .MFValueFunction import MFValueFunction


class UCBLearner(MFValueFunction):
    def __init__(self, stop=14, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.stop = stop


    def reset(self, observation):
        train_dataset = observation
        super().reset(train_dataset)
        self.train_dataset = train_dataset
        self.train_consumption_matrix = scipy.sparse.csr_matrix(
            (self.train_dataset.data[:, 2],
             (self.train_dataset.data[:, 0], self.train_dataset.data[:, 1])),
            (self.train_dataset.num_total_users,
             self.train_dataset.num_total_items))
        self.num_total_items = self.train_dataset.num_total_items

        items_entropy = Entropy.get_items_entropy(
            self.train_consumption_matrix)
        items_popularity = MostPopular.get_items_popularity(
            self.train_consumption_matrix, normalize=False)
        self.items_bias = LogPopEnt.get_items_logpopent(
            items_popularity, items_entropy)

        mf_model = mf.SVD(num_lat=self.num_lat)
        mf_model.fit(self.train_consumption_matrix)
        self.items_weights = mf_model.items_weights
        self.num_latent_factors = len(self.items_weights[0])

        self.I = np.eye(len(self.items_weights[0]))
        self.bs = defaultdict(lambda: np.zeros(self.num_latent_factors))
        self.As = defaultdict(lambda: self.I.copy())
        self.users_nb_items = defaultdict(lambda: 0)

    @staticmethod
    def discount_bias(num_total_items, stop):
        limit = pow(2, stop) / 100
        return pow(2, min(stop, num_total_items)) / limit

    def action_estimates(self, candidate_actions):
        uid = candidate_actions[0]
        candidate_items = candidate_actions[1]
        b = self.bs[uid]
        A = self.As[uid]
        items_bias = self.items_bias
        mean = np.dot(np.linalg.inv(A), b)
        pred_rule = mean @ self.items_weights[user_candidate_items].T
        current_bias = items_bias[user_candidate_items] * max(
            1, np.max(pred_rule))
        bias = current_bias - (current_bias * self.discount_bias(
            self.users_nb_items[uid], self.stop) / 100)
        bias[bias < 0] = 0
        items_score = (pred_rule + bias)[::-1]
        return items_score

    def update(self, observation, action, reward, info):
        uid = action[0]
        item = action[1]
        additional_data = info
        b = self.bs[uid]
        A = self.As[uid]
        max_item_weight = self.items_weights[item]
        A += max_item_weight[:, None].dot(max_item_weight[None, :])
        b += reward * max_item_weight
        self.users_nb_items[uid] += 1



import numpy as np
import scipy
import scipy.stats
import sys, os
import random
from threadpoolctl import threadpool_limits
import ctypes

from irec.utils.utils import run_parallel
from . import MF


class ICFPMF(MF):
    def __init__(self,
                 iterations=50,
                 var=1,
                 user_var=1,
                 item_var=1,
                 *args,
                 **kwargs):
        super().__init__(*args, **kwargs)
        self.iterations = iterations
        self.var = var
        self.user_var = user_var
        self.item_var = item_var
        self.user_lambda = self.var / self.user_var
        self.item_lambda = self.var / self.item_var
        self.objective_values = []
        self.best = None


    def load_var(self, training_matrix):
        decimals = 4
        training_matrix = self.normalize_matrix(training_matrix)
        self.var = np.var(training_matrix)
        # self.user_var = self.var
        # self.item_var = self.var
        self.user_var = np.mean(np.var(training_matrix, axis=1))
        self.item_var = np.mean(np.var(training_matrix, axis=0))
        self.user_lambda = self.var / self.user_var
        self.item_lambda = self.var / self.item_var
        self.var = np.round(self.var, decimals)
        self.user_var = np.round(self.user_var, decimals)
        self.item_var = np.round(self.item_var, decimals)
        self.user_lambda = np.round(self.user_lambda, decimals)
        self.item_lambda = np.round(self.item_lambda, decimals)

    def fit(self, training_matrix):
        super().fit()
        training_matrix = self.normalize_matrix(training_matrix)
        self_id = id(self)
        self.training_matrix = training_matrix
        num_users = training_matrix.shape[0]
        num_items = training_matrix.shape[1]
        self.lowest_value = lowest_value = np.min(training_matrix)
        highest_value = np.max(training_matrix)
        self.observed_ui = observed_ui = np.nonzero(
            training_matrix > lowest_value)  # itens observed by some user
        self.I = I = np.eye(self.num_lat)
        self.users_weights = np.random.multivariate_normal(
            np.zeros(self.num_lat), self.user_var * I,
            training_matrix.shape[0])
        self.items_weights = np.random.multivariate_normal(
            np.zeros(self.num_lat), self.item_var * I,
            training_matrix.shape[1])
        best_objective_value = np.inf
        # without burning
        np.seterr('warn')
        for i in range(self.iterations):
            print(f'[{i+1}/{self.iterations}]')
            with threadpool_limits(limits=1, user_api='blas'):
                for to_run in random.sample([1, 2], 2):
                    if to_run == 1:
                        self.users_means = np.zeros((num_users, self.num_lat))
                        self.users_covs = np.zeros(
                            (num_users, self.num_lat, self.num_lat))
                        args = [(
                            self_id,
                            i,
                        ) for i in range(num_users)]
                        results = run_parallel(self.compute_user_weight, args)
                        for uid, (mean, cov, weight) in enumerate(results):
                            self.users_means[uid] = mean
                            self.users_covs[uid] = cov
                            self.users_weights[uid] = weight
                    else:
                        self.items_means = np.zeros((num_items, self.num_lat))
                        self.items_covs = np.zeros(
                            (num_items, self.num_lat, self.num_lat))
                        args = [(
                            self_id,
                            i,
                        ) for i in range(num_items)]
                        results = run_parallel(self.compute_item_weight, args)
                        for iid, (mean, cov, weight) in enumerate(results):
                            self.items_means[iid] = mean
                            self.items_covs[iid] = cov
                            self.items_weights[iid] = weight

            rmse = np.sqrt(
                np.mean((self.get_predicted()[observed_ui] -
                         training_matrix[observed_ui])**2))
            # map_value = scipy.special.logsumexp(scipy.stats.norm.pdf(training_matrix[observed_ui],(self.users_weights @ self.items_weights.T)[observed_ui],self.var))\
            #         * scipy.special.logsumexp([scipy.stats.multivariate_normal.pdf(weights,means,covs)
            #                                for means, covs, weights in zip(self.users_means,self.users_covs,self.users_weights)])\
            #         * scipy.special.logsumexp([scipy.stats.multivariate_normal.pdf(weights,means,covs)
            #                                for means, covs, weights in zip(self.items_means,self.items_covs,self.items_weights)])

            objective_value = np.sum((training_matrix[observed_ui] - self.get_predicted()[observed_ui])**2)/2 +\
                self.user_lambda/2 * np.sum(np.linalg.norm(self.users_weights,axis=1)**2) +\
                self.item_lambda/2 * np.sum(np.linalg.norm(self.items_weights,axis=1)**2)

            print("Objective value", objective_value)
            #     self.objective_values.append(objective_value)
            print("RMSE", rmse)

            # if self.best == None:
            #     self.best = self.__deepcopy__()
            #     best_objective_value = objective_value
            # else:
            #     if objective_value < best_objective_value:
            #         self.best = self.__deepcopy__()
            #         best_objective_value = objective_value

        # self = self.best
        # del self.best
        del self.training_matrix
        del self.lowest_value

    @staticmethod
    def compute_user_weight(obj_id, uid):
        self = ctypes.cast(obj_id, ctypes.py_object).value
        training_matrix = self.training_matrix
        lowest_value = self.lowest_value
        I = self.I
        observed = training_matrix[uid, :] > lowest_value
        tmp = np.linalg.inv(
            (np.dot(self.items_weights[observed].T,
                    self.items_weights[observed]) + I * self.user_lambda))
        mean = tmp.dot(self.items_weights[observed].T).dot(
            training_matrix[uid, observed])
        cov = tmp * self.var
        return mean, cov, np.random.multivariate_normal(mean, cov)

    @staticmethod
    def compute_item_weight(obj_id, iid):
        self = ctypes.cast(obj_id, ctypes.py_object).value
        training_matrix = self.training_matrix
        lowest_value = self.lowest_value
        I = self.I
        observed = training_matrix[:, iid] > lowest_value
        tmp = np.linalg.inv(
            (np.dot(self.users_weights[observed].T,
                    self.users_weights[observed]) + I * self.item_lambda))
        mean = tmp.dot(self.users_weights[observed].T).dot(
            training_matrix[observed, iid])
        cov = tmp * self.var
        return mean, cov, np.random.multivariate_normal(mean, cov)

    def __deepcopy__(self):
        new = type(self)()
        new.__dict__.update(self.__dict__)
        new.users_weights = self.users_weights.copy()
        new.users_means = self.users_means.copy()
        new.users_covs = self.users_covs.copy()
        new.items_weights = self.items_weights.copy()
        new.items_means = self.items_means.copy()
        new.items_covs = self.items_covs.copy()
        return new



import numpy as np
from tqdm import tqdm
from . import ValueFunction, Entropy, MostPopular, LogPopEnt, ExperimentalValueFunction
import matplotlib.pyplot as plt
import scipy.stats
import os


class HELF(ExperimentalValueFunction):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    @staticmethod
    def get_items_helf(items_popularity, items_entropy, num_total_users):

        a = np.ma.log(items_popularity).filled(0) / np.log(num_total_users)
        b = items_entropy / np.max(items_entropy)
        print(np.sort(b))
        print(b.min())
        print(a.min())
        print(np.sort(a))
        np.seterr('warn')
        items_helf = 2 * a * b / (a + b)
        items_helf[np.isnan(items_helf)] = 0
        return items_helf

    def reset(self, observation):
        train_dataset = observation
        super().reset(train_dataset)
        self.train_dataset = train_dataset
        self.train_consumption_matrix = scipy.sparse.csr_matrix(
            (self.train_dataset.data[:, 2],
             (self.train_dataset.data[:, 0], self.train_dataset.data[:, 1])),
            (self.train_dataset.num_total_users,
             self.train_dataset.num_total_items))
        self.num_total_items = self.train_dataset.num_total_items
        # mask = np.ones(self.train_consumption_matrix.shape[0], dtype=bool)
        # mask[uids] = 0
        num_train_users = len(self.train_dataset.uids)
        items_entropy = Entropy.get_items_entropy(
            self.train_consumption_matrix)
        items_popularity = MostPopular.get_items_popularity(
            self.train_consumption_matrix, normalize=False)
        self.items_logpopent = HELF.get_items_helf(items_popularity,
                                                   items_entropy,
                                                   num_train_users)

    def action_estimates(self, candidate_actions):
        uid = candidate_actions[0]
        candidate_items = candidate_actions[1]
        items_score = self.items_logpopent[candidate_items]
        return items_score, None
        # top_items = list(reversed(np.argsort(items_score)))[:self.interaction_size]

        # top_iids = list(reversed(np.argsort(items_logpopent)))[:self.get_iterations()]
        # num_total_users = len(uids)

        # fig, ax = plt.subplots()
        # ax.scatter(items_entropy,items_popularity,marker="D",color='darkblue')
        # ax.set_ylabel("Popularity")
        # ax.set_xlabel("Entropy")
        # for start, end, color in [(0,10,'green'),(10,20,'red'),(20,30,'darkred'),(30,40,'yellow'),(40,50,'orange')]:
        #     ax.scatter(items_entropy[top_iids[start:end]],items_popularity[top_iids[start:end]],marker='D',color=color)
        # fig.savefig(os.path.join(self.DIRS['img'],"corr_popent_"+self.get_id()+".png"))

        # for idx_uid in tqdm(range(num_total_users)):
        #     uid = uids[idx_uid]
        #     self.results[uid].extend(top_iids)
        # self.save_results()



import numpy as np
from tqdm import tqdm
from . import ValueFunction, Entropy, MostPopular, LogPopEnt, PopPlusEnt, ExperimentalValueFunction
import matplotlib.pyplot as plt
import scipy.stats
import os


class PPELPE(ExperimentalValueFunction):
    @staticmethod
    def get_items_ppelpe(items_popularity, items_entropy, do_sum=True):
        items_popplusent = PopPlusEnt.get_items_popplusent(
            items_popularity, items_entropy)
        items_logpopent = LogPopEnt.get_items_logpopent(
            items_popularity, items_entropy)
        if do_sum:
            res = items_popplusent + items_logpopent
        else:
            res = items_popplusent * items_logpopent
        return res / np.max(res)

    def reset(self, observation):
        train_dataset = observation
        super().reset(train_dataset)

        items_entropy = Entropy.get_items_entropy(
            self.train_consumption_matrix)
        items_popularity = MostPopular.get_items_popularity(
            self.train_consumption_matrix, normalize=False)
        self.items_ppelpe = self.get_items_ppelpe(items_popularity,
                                                  items_entropy)

    def action_estimates(self, candidate_actions):
        uid = candidate_actions[0]
        candidate_items = candidate_actions[1]
        items_score = self.items_ppelpe[candidate_items]
        return items_score, None

        # correlation = scipy.stats.pearsonr(items_entropy,items_popularity)[0]

        # top_iids = list(reversed(np.argsort(items_ppelpe)))[:self.get_iterations()]

        # fig, ax = plt.subplots()
        # ax.scatter(items_entropy,items_popularity,marker="D",color='darkblue')
        # ax.set_ylabel("Popularity")
        # ax.set_xlabel("Entropy")
        # ax.text(0.3, 0.9 , f'Correlation coefficient: {correlation:.2f}', color='k',
        #         ha='center', va='center',
        #         bbox=dict(facecolor='none', edgecolor='k', pad=10.0),
        #         transform = ax.transAxes)
        # for start, end, color in [(0,10,'green'),(10,20,'red'),(20,30,'darkred'),(30,40,'yellow'),(40,50,'orange')]:
        #     print(top_iids[start:end])
        #     ax.scatter(items_entropy[top_iids[start:end]],items_popularity[top_iids[start:end]],marker='D',color=color)
        #     print("[%d,%d] sum(popularity)=%.2f sum(entropy)=%.2f"%(start,end,
        #                                                          np.sum(items_popularity[top_iids[start:end]]/np.max(items_popularity)),
        #                                                          np.sum(items_entropy[top_iids[start:end]]/np.max(items_entropy))))
        # fig.savefig(os.path.join(self.DIRS['img'],"corr_popent_"+self.get_id()+".png"))

        # num_total_users = len(uids)
        # for idx_uid in tqdm(range(num_total_users)):
        #     uid = uids[idx_uid]
        #     self.results[uid].extend(top_iids)
        # self.save_results()



import numpy as np
from tqdm import tqdm
from .ExperimentalValueFunction import ExperimentalValueFunction
from .MostPopular import *
import matplotlib.pyplot as plt
import os
import scipy
from collections import defaultdict
import random


class MostPopular(ExperimentalValueFunction):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    @staticmethod
    def get_items_popularity(consumption_matrix, normalize=True):
        lowest_value = np.min(consumption_matrix)
        if not isinstance(consumption_matrix, scipy.sparse.spmatrix):
            items_popularity = np.count_nonzero(
                consumption_matrix > lowest_value, axis=0)
        else:
            items_popularity = np.array(
                np.sum(consumption_matrix > lowest_value, axis=0)).flatten()

        if normalize:
            items_popularity = items_popularity / consumption_matrix.shape[0]

        return items_popularity

    def reset(self, observation):
        train_dataset = observation
        super().reset(train_dataset)
        self.train_dataset = train_dataset
        self.num_total_items = self.train_dataset.num_total_items
        self.train_consumption_matrix = scipy.sparse.csr_matrix(
            (self.train_dataset.data[:, 2],
             (self.train_dataset.data[:, 0], self.train_dataset.data[:, 1])),
            (self.train_dataset.num_total_users,
             self.train_dataset.num_total_items))
        self.null_value = np.min(self.train_consumption_matrix)
        self.items_popularity = self.get_items_popularity(
            self.train_consumption_matrix, normalize=False)

    def action_estimates(self, candidate_actions):
        uid = candidate_actions[0]
        candidate_items = candidate_actions[1]
        items_score = self.items_popularity[candidate_items]
        return items_score, None

    def update(self, observation, action, reward, info):
        uid = action[0]
        item = action[1]
        additional_data = info
        if reward != self.null_value:
            self.items_popularity[item] += 1



import numpy as np
from tqdm import tqdm
from .ExperimentalValueFunction import ExperimentalValueFunction
from .BestRated import *
import matplotlib.pyplot as plt
import os
import scipy
from collections import defaultdict
import random


class BestRated(ExperimentalValueFunction):
    """Best Rated.
    
    Recommends top-rated items based on their average ratings in each iteration.    
    """
    def __init__(self, *args, **kwargs):
        """__init__.

        Args:
            args:
            kwargs:
        """
        super().__init__(*args, **kwargs)

    @staticmethod
    def get_items_bestrated(consumption_matrix):
        """get_items_bestrated.
        
        Args:
            consumption_matrix: consumption_matrix
            normalize (bool): normalize

        Returns:
            numpy.ndarray:
        """

        items_bestrated = np.array(np.mean(consumption_matrix, axis=0)).flatten()
        return items_bestrated

    def reset(self, observation):
        """reset.

        Args:
            observation: 
        """
        train_dataset = observation
        super().reset(train_dataset)
        self.train_dataset = train_dataset
        self.num_total_items = self.train_dataset.num_total_items
        self.train_consumption_matrix = scipy.sparse.csr_matrix(
            (self.train_dataset.data[:, 2],
             (self.train_dataset.data[:, 0], self.train_dataset.data[:, 1])),
            (self.train_dataset.num_total_users,
             self.train_dataset.num_total_items))
        self.null_value = np.min(self.train_consumption_matrix)
        self.items_bestrated = self.get_items_bestrated(
            self.train_consumption_matrix)

    def action_estimates(self, candidate_actions):
        """action_estimates.

        Args:
            candidate_actions: (user id, candidate_items)
        
        Returns:
            numpy.ndarray:
        """
        uid = candidate_actions[0]
        candidate_items = candidate_actions[1]
        items_score = self.items_bestrated[candidate_items]

        return items_score, None

    def update(self, observation, action, reward, info):
        """update.

        Args:
            observation:
            action: (user id, item)
            reward (float): reward
            info: 
        """
        uid = action[0]
        item = action[1]
        additional_data = info
        if reward != self.null_value:
            self.items_bestrated[item] = self.train_consumption_matrix[:,item].toarray().mean()



import numpy as np
from tqdm import tqdm
from .ExperimentalValueFunction import ExperimentalValueFunction
import matplotlib.pyplot as plt
import os
import scipy.sparse
from collections import defaultdict
import random
import itertools
import random


class kNNBandit(ExperimentalValueFunction):
    """k Nearest-Neighbors Bandit.
    
    A simple multi-armed bandit elaboration of neighbor-based collaborative filtering. 
    A variant of the nearest-neighbors scheme, but endowed with a controlled stochastic exploration 
    capability of the users’ neighborhood, by a parameter-free application of Thompson sampling [1]_.

    References
    ----------
    .. [1] Sanz-Cruzado, Javier, Pablo Castells, and Esther López. "A simple multi-armed nearest-neighbor 
       bandit for interactive recommendation." Proceedings of the 13th ACM Conference on Recommender Systems. 2019.
    """
    def __init__(self, alpha_0, beta_0, k, *args, **kwargs):
        """__init__.

        Args:
            args:
            kwargs:
        """
        super().__init__(*args, **kwargs)
        self.alpha_0 = alpha_0
        self.beta_0 = beta_0
        self.k = k


    def reset(self, observation):
        """reset.

        Args:
            observation: 
        """ 
        train_dataset = observation
        super().reset(train_dataset)
        self.train_dataset = train_dataset
        self.train_consumption_matrix = scipy.sparse.csc_matrix(
            (self.train_dataset.data[:, 2],
             (self.train_dataset.data[:, 0], self.train_dataset.data[:, 1])),
            (self.train_dataset.num_total_users,
             self.train_dataset.num_total_items))
        self.train_consumption_matrix = (self.train_consumption_matrix >=
                                         4).astype(int)

        self.num_total_users = self.train_dataset.num_total_users
        self.num_total_items = self.train_dataset.num_total_items
        self.consumption_matrix = self.train_consumption_matrix.tolil()

        self.users_alphas = np.zeros(
            (self.num_total_users, self.num_total_users), dtype='float32')
        self.users_rating_sum = np.zeros(
            (self.num_total_users)) + self.alpha_0 + self.beta_0
        l_ = range(len(self.train_dataset.data))
        self.items_consumed_users = defaultdict(list)
        self.items_consumed_users_ratings = defaultdict(list)
        for i in tqdm(l_):
            uid = int(self.train_dataset.data[i, 0])
            item = int(self.train_dataset.data[i, 1])
            reward = self.train_dataset.data[i, 2]
            reward = reward >= 4
            self.users_rating_sum[uid] += reward
            if len(self.items_consumed_users[item]) > 0:
                item_consumed_uids = np.array(
                    [i for i in self.items_consumed_users[item]])
                item_ratings = np.array(
                    [i for i in self.items_consumed_users_ratings[item]])
                self.users_alphas[uid,
                                  item_consumed_uids] += reward * item_ratings
            self.items_consumed_users[item].append(uid)
            self.items_consumed_users_ratings[item].append(reward)

        del self.train_consumption_matrix

    def action_estimates(self, candidate_actions):
        """action_estimates.

        Args:
            candidate_actions: (user id, candidate_items)
        
        Returns:
            numpy.ndarray:
        """
        uid = candidate_actions[0]
        candidate_items = candidate_actions[1]
        users_score = np.zeros(self.num_total_users - 1)
        uids = np.array(list(set(range(self.num_total_users)) - {uid}))
        vs1 = self.users_alphas[uid, uids]
        vs2 = self.users_rating_sum[uids] - self.users_alphas[uid, uids]
        for i, uid_ in enumerate(uids):
            v1 = vs1[i]
            v2 = vs2[i]
            users_score[i] = np.random.beta(v1 + self.alpha_0,
                                            v2 + self.beta_0)

        idxs = np.argpartition(users_score, -self.k)[-self.k:]
        top_uids = uids[idxs]

        top_users_score = users_score[idxs]

        items_score = np.zeros(len(candidate_items))
        for top_user_score, top_uid in zip(top_users_score, top_uids):
            items_score += top_user_score * self.consumption_matrix[
                top_uid].A.flatten()[candidate_items]

        idxs = np.where(np.max(items_score) == items_score)[0]
        items_score[idxs] += np.random.rand(len(idxs))

        return items_score, None

    def update(self, observation, action, reward, info):
        """update.

        Args:
            observation:
            action: (user id, item)
            reward (float): reward
            info: 
        """
        uid = action[0]
        item = action[1]
        additional_data = info
        reward = reward >= 4

        if len(self.items_consumed_users[item]) > 0:
            item_consumed_uids = np.array(
                [i for i in self.items_consumed_users[item]])
            item_ratings = np.array(
                [i for i in self.items_consumed_users_ratings[item]])
            self.users_alphas[uid, item_consumed_uids] += reward * item_ratings

        self.users_rating_sum[uid] += reward
        self.consumption_matrix[uid, item] = reward
        self.items_consumed_users[item].append(uid)
        self.items_consumed_users_ratings[item].append(reward)



from collections import defaultdict


class Recommender:
    """Recommender."""

    def __init__(self, result_list_size=10, *args, **kwargs):
        """__init__.

        Args:
            result_list_size:
            args:
            kwargs:
        """
        super().__init__(*args, **kwargs)
        self.result_list_size = result_list_size
        self.results = defaultdict(list)

    def train(self):
        """train."""
        ṕass

    def predict(self):
        """predict."""
        self.results.clear()



from .ICF import ICF
import numpy as np
from tqdm import tqdm
from threadpoolctl import threadpool_limits
import ctypes
from collections import defaultdict
import joblib
import scipy
import mf
from .LinearICF import LinearICF
from .LinearUCB import LinearUCB


class LinearUCB1(LinearUCB):
    def reset(self, observation):
        train_dataset = observation
        super().reset(train_dataset)
        self.bs = defaultdict(lambda: np.ones(self.num_latent_factors))



from .ExperimentalValueFunction import ExperimentalValueFunction
import numpy as np
from tqdm import tqdm
#import util
from threadpoolctl import threadpool_limits
import scipy
import scipy.stats
import ctypes
import mf
from collections import defaultdict
import value_functions
from .MFValueFunction import MFValueFunction
from .LinUCB import LinUCB


def _prediction_rule(A, b, items_weights, alpha):
    mean = np.dot(np.linalg.inv(A), b)
    items_uncertainty = alpha * np.sqrt(
        np.sum(items_weights.dot(np.linalg.inv(A)) * items_weights, axis=1))
    items_user_similarity = mean @ items_weights.T
    items_score = items_user_similarity + items_uncertainty
    return items_score


class ICLinUCB(LinUCB):
    def __init__(self, u,num_clusters, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.u= u
        self.num_clusters = num_clusters


    def reset(self, observation):
        train_dataset = observation
        super().reset(train_dataset)
        self.train_consumption_matrix

        kmeans = KMeans(self.num_clusters).fit(self.train_consumption_matrix.T)
        self.groups = list(range(self.num_clusters))

        self.groups_items = defaultdict(list)
        for uid, group in enumerate(kmeans.labels_):
            self.groups_items[group].append(uid)


    def action_estimates(self, candidate_actions):
        self.action_estimates(candidate_actions)
        return items_score, None

    def update(self, observation, action, reward, info):
        self.update(observation, action, reward, info)
        uid = action[0]
        item = action[1]
        additional_data = info
        max_item_latent_factors = self.items_weights[item]
        b = self.bs[uid]
        A = self.As[uid]
        A += max_item_latent_factors[:, None].dot(
            max_item_latent_factors[None, :])
        b += reward * max_item_latent_factors



import numpy as np

import irec.value_functions

import scipy.sparse
from collections import defaultdict
import time
from irec.utils.utils import run_parallel
import ctypes

# import irec.utils.dataset as dataset
from irec.utils import dataset
from irec.value_functions.Entropy import Entropy
from irec.RelevanceEvaluator import ThresholdRelevanceEvaluator
from irec.metrics import ILD, Recall, Precision, EPC, EPD

np.seterr(all="raise")


class MetricEvaluator:
    """MetricsEvaluator."""

    def __init__(self, relevance_evaluator_threshold: float, *args, **kwargs):
        """__init__.

        Args:
            relevance_evaluator_threshold (float): relevance_evaluator_threshold
            args:
            kwargs:
        """
        super().__init__(*args, **kwargs)
        self.relevance_evaluator = ThresholdRelevanceEvaluator(
            relevance_evaluator_threshold
        )


class TotalMetricEvaluator(MetricEvaluator):
    def __init__(self, ground_truth_dataset, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.ground_truth_dataset = ground_truth_dataset
        if isinstance(ground_truth_dataset, dataset.Dataset):
            self.ground_truth_consumption_matrix = scipy.sparse.csr_matrix(
                (
                    self.ground_truth_dataset.data[:, 2],
                    (
                        self.ground_truth_dataset.data[:, 0],
                        self.ground_truth_dataset.data[:, 1],
                    ),
                ),
                (
                    self.ground_truth_dataset.num_total_users,
                    self.ground_truth_dataset.num_total_items,
                ),
            )

    def _metric_evaluation(self, metric_class):
        start_time = time.time()
        metric_values = []
        if issubclass(metric_class, Recall):
            metric = metric_class(
                users_false_negative=self.users_false_negative,
                ground_truth_dataset=self.ground_truth_dataset,
                relevance_evaluator=self.relevance_evaluator,
            )
        else:
            metric = metric_class(
                ground_truth_dataset=self.ground_truth_dataset,
                relevance_evaluator=self.relevance_evaluator,
            )
        start_time = time.time()
        i = 0
        while i < len(self.results):
            uid = self.results[i][0]
            item = self.results[i][1]
            metric.update_recommendation(
                uid, item, self.ground_truth_consumption_matrix[uid, item]
            )
            i += 1

        metric_values.append([metric.compute(uid) for uid in self.uids])

        print(
            f"{self.__class__.__name__} spent {time.time()-start_time:.2f} seconds executing {metric_class.__name__} metric"
        )
        return metric_values

    def evaluate(self, metric_class, results):
        self.users_false_negative = defaultdict(int)
        for row in self.ground_truth_dataset.data:
            uid = int(row[0])
            reward = row[2]
            if self.relevance_evaluator.is_relevant(reward):
                self.users_false_negative[uid] += 1
        uids = []
        for uid, _ in results:
            uids.append(uid)
        uids = list(set(uids))
        self.uids = uids
        self.results = results

        metric_values = self._metric_evaluation(metric_class)

        return metric_values


class CumulativeMetricEvaluator(MetricEvaluator):
    def __init__(self, ground_truth_dataset, buffer_size, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if ground_truth_dataset != None:
            self.ground_truth_dataset = ground_truth_dataset
            self.ground_truth_consumption_matrix = scipy.sparse.csr_matrix(
                (
                    self.ground_truth_dataset.data[:, 2],
                    (
                        self.ground_truth_dataset.data[:, 0],
                        self.ground_truth_dataset.data[:, 1],
                    ),
                ),
                (
                    self.ground_truth_dataset.num_total_users,
                    self.ground_truth_dataset.num_total_items,
                ),
            )
        self.buffer_size = buffer_size

    def _metric_evaluation(self, metric_class):
        if issubclass(metric_class, Recall):
            metric = metric_class(
                users_false_negative=self.users_false_negative,
                ground_truth_dataset=self.ground_truth_dataset,
                relevance_evaluator=self.relevance_evaluator,
            )
        else:
            metric = metric_class(
                ground_truth_dataset=self.ground_truth_dataset,
                relevance_evaluator=self.relevance_evaluator,
            )

        start = 0
        start_time = time.time()
        metric_values = []
        while start < len(self.results):
            for i in range(start, min(start + self.buffer_size, len(self.results))):
                uid = self.results[i][0]
                item = self.results[i][1]
                metric.update_recommendation(
                    uid, item, self.ground_truth_consumption_matrix[uid, item]
                )
            start = min(start + self.buffer_size, len(self.results))
            metric_values.append(np.mean([metric.compute(uid) for uid in self.uids]))
        print(
            f"{self.__class__.__name__} spent {time.time()-start_time:.2f} seconds executing {metric_class.__name__} metric"
        )
        return metric_values

    def evaluate(self, metric_class, results):
        self.users_false_negative = defaultdict(int)
        for row in self.ground_truth_dataset.data:
            uid = int(row[0])
            reward = row[2]
            if self.relevance_evaluator.is_relevant(reward):
                self.users_false_negative[uid] += 1
        uids = []
        for uid, _ in results:
            uids.append(uid)
        uids = list(set(uids))
        self.uids = uids
        self.results = results
        metric_values = self._metric_evaluation(metric_class)
        return metric_values


class InteractionMetricEvaluator(MetricEvaluator):
    def __init__(
        self,
        ground_truth_dataset,
        num_interactions,
        interaction_size,
        interactions_to_evaluate=None,
        *args,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        self.ground_truth_dataset = ground_truth_dataset
        self.num_interactions = num_interactions
        self.interaction_size = interaction_size
        self.interactions_to_evaluate = interactions_to_evaluate
        if self.interactions_to_evaluate == None:
            self.interactions_to_evaluate = list(range(self.num_interactions))
        self.iterations_to_evaluate = self.interactions_to_evaluate

        if isinstance(ground_truth_dataset, dataset.Dataset):
            self.ground_truth_consumption_matrix = scipy.sparse.csr_matrix(
                (
                    self.ground_truth_dataset.data[:, 2],
                    (
                        self.ground_truth_dataset.data[:, 0],
                        self.ground_truth_dataset.data[:, 1],
                    ),
                ),
                # (
                # np.max(self.ground_truth_dataset.uids),
                # np.max(self.ground_truth_dataset.iids),
                # self.ground_truth_dataset.num_total_items,
                # ),
                (
                    self.ground_truth_dataset.num_total_users,
                    self.ground_truth_dataset.num_total_items,
                ),
            )

    def _metric_evaluation(self, metric_class):
        start_time = time.time()
        metric_values = []
        for i in range(self.num_interactions):
            if issubclass(metric_class, Recall):
                metric = metric_class(
                    users_false_negative=self.users_false_negative,
                    ground_truth_dataset=self.ground_truth_dataset,
                    relevance_evaluator=self.relevance_evaluator,
                )
            elif issubclass(metric_class, ILD):
                metric = metric_class(
                    items_distance=self.items_distance,
                    ground_truth_dataset=self.ground_truth_dataset,
                    relevance_evaluator=self.relevance_evaluator,
                )
            elif issubclass(metric_class, EPC):
                metric = metric_class(
                    items_normalized_popularity=self.items_normalized_popularity,
                    ground_truth_dataset=self.ground_truth_dataset,
                    relevance_evaluator=self.relevance_evaluator,
                )
            else:
                metric = metric_class(
                    ground_truth_dataset=self.ground_truth_dataset,
                    relevance_evaluator=self.relevance_evaluator,
                )
            for uid in self.uids:
                interaction_results = self.users_items_recommended[uid][
                    i * self.interaction_size : i * self.interaction_size
                    + self.interaction_size
                ]
                for item in interaction_results:
                    metric.update_recommendation(
                        uid, item, self.ground_truth_consumption_matrix[uid, item]
                    )

            metric_values.append(np.mean([metric.compute(uid) for uid in self.uids]))

        print(
            f"{self.__class__.__name__} spent {time.time()-start_time:.2f} seconds executing {metric_class.__name__} metric"
        )
        return metric_values

    def evaluate(
        self,
        metric_class,
        results,
    ):
        self.users_false_negative = defaultdict(int)
        for row in self.ground_truth_dataset.data:
            uid = int(row[0])
            reward = row[2]
            if self.relevance_evaluator.is_relevant(reward):
                self.users_false_negative[uid] += 1

        self.users_items_recommended = defaultdict(list)
        for uid, item in results:
            self.users_items_recommended[uid].append(item)
        self.uids = list(self.users_items_recommended.keys())
        if isinstance(metric_class, ILD):
            self.items_distance = get_items_distance(
                self.ground_truth_consumption_matrix
            )
        if isinstance(metric_class, EPC):
            self.items_normalized_popularity = (
                irec.value_functions.MostPopular.get_items_popularity(
                    self.ground_truth_consumption_matrix
                )
            )
        if isinstance(metric_class, Entropy):
            self.items_entropy = irec.value_functions.Entropy.get_items_entropy(
                self.ground_truth_consumption_matrix
            )

        metric_values = self._metric_evaluation(metric_class)

        return metric_values


class CumulativeInteractionMetricEvaluator(InteractionMetricEvaluator):
    @staticmethod
    def metric_summarize(users_metric_values):
        return np.mean(list(users_metric_values.values()))

    def _metric_evaluation(self, metric_class):
        start_time = time.time()
        metric_values = []
        if issubclass(metric_class, Recall):
            metric = metric_class(
                users_false_negative=self.users_false_negative,
                ground_truth_dataset=self.ground_truth_dataset,
                relevance_evaluator=self.relevance_evaluator,
            )
        elif issubclass(metric_class, ILD):
            metric = metric_class(
                items_distance=self.items_distance,
                ground_truth_dataset=self.ground_truth_dataset,
                relevance_evaluator=self.relevance_evaluator,
            )
        elif issubclass(metric_class, EPC):
            metric = metric_class(
                items_normalized_popularity=self.items_normalized_popularity,
                ground_truth_dataset=self.ground_truth_dataset,
                relevance_evaluator=self.relevance_evaluator,
            )
        elif issubclass(metric_class, Entropy):
            metric = metric_class(
                items_entropy=self.items_entropy,
                ground_truth_dataset=self.ground_truth_dataset,
                relevance_evaluator=self.relevance_evaluator,
            )
        else:
            metric = metric_class(
                ground_truth_dataset=self.ground_truth_dataset,
                relevance_evaluator=self.relevance_evaluator,
            )
        for i in range(self.num_interactions):
            for uid in self.uids:
                interaction_results = self.users_items_recommended[uid][
                    i * self.interaction_size : i * self.interaction_size
                    + self.interaction_size
                ]
                for item in interaction_results:
                    metric.update_recommendation(
                        uid, item, self.ground_truth_consumption_matrix[uid, item]
                    )

            if (i + 1) in self.interactions_to_evaluate:
                print(f"Computing interaction {i+1} with {self.__class__.__name__}")
                metric_values.append(
                    self.metric_summarize(
                        {uid: metric.compute(uid) for uid in self.uids}
                    )
                )

        print(
            f"{self.__class__.__name__} spent {time.time()-start_time:.2f} seconds executing {metric_class.__name__} metric"
        )
        return metric_values


class UserCumulativeInteractionMetricEvaluator(CumulativeInteractionMetricEvaluator):
    @staticmethod
    def metric_summarize(users_metric_values):
        return users_metric_values


class IterationsMetricEvaluator(InteractionMetricEvaluator):
    @staticmethod
    def metric_summarize(users_metric_values):
        return np.mean(users_metric_values)

    def _metric_evaluation(self, metric_class):
        start_time = time.time()
        metric_values = []
        if issubclass(metric_class, Recall):
            metric = metric_class(
                users_false_negative=self.users_false_negative,
                ground_truth_dataset=self.ground_truth_dataset,
                relevance_evaluator=self.relevance_evaluator,
            )
        elif issubclass(metric_class, ILD):
            metric = metric_class(
                items_distance=self.items_distance,
                ground_truth_dataset=self.ground_truth_dataset,
                relevance_evaluator=self.relevance_evaluator,
            )
        elif issubclass(metric_class, EPC):
            metric = metric_class(
                items_normalized_popularity=self.items_normalized_popularity,
                ground_truth_dataset=self.ground_truth_dataset,
                relevance_evaluator=self.relevance_evaluator,
            )
        else:
            metric = metric_class(
                ground_truth_dataset=self.ground_truth_dataset,
                relevance_evaluator=self.relevance_evaluator,
            )
        if 0 not in self.iterations_to_evaluate:
            self.iterations_to_evaluate = [0] + self.iterations_to_evaluate
        for i in range(len(self.iterations_to_evaluate) - 1):
            for uid in self.uids:
                interaction_results = self.users_items_recommended[uid][
                    self.iterations_to_evaluate[i] : self.iterations_to_evaluate[i + 1]
                ]
                for item in interaction_results:
                    metric.update_recommendation(
                        uid, item, self.ground_truth_consumption_matrix[uid, item]
                    )
            # if (i+1) in interactions_to_evaluate:
            print(
                f"Computing iteration {self.iterations_to_evaluate[i+1]} with {self.__class__.__name__}"
            )
            metric_values.append(
                self.metric_summarize([metric.compute(uid) for uid in self.uids])
            )

        print(
            f"{self.__class__.__name__} spent {time.time()-start_time:.2f} seconds executing {metric_class.__name__} metric"
        )
        return metric_values



from . import MF
import sklearn.decomposition


class NMF(MF):
    def fit(self, training_matrix):
        model = sklearn.decomposition.NMF(n_components=self.num_lat,
                                          init='nndsvd',
                                          random_state=0)
        self.users_weights = model.fit_transform(training_matrix)
        self.items_weights = model.components_.T



from .ICF import *
from .LinEGreedy import *
import numpy as np
import random
from tqdm import tqdm
from threadpoolctl import threadpool_limits
import ctypes
import scipy
import joblib
from .LinearICF import LinearICF


class LinearEGreedy(LinearICF):
    """Linear Epsilon Greedy.
    
    A linear exploitation of the items latent factors defined by a PMF
    formulation that also explore random items with probability ε [1]_.

    References
    ----------
    .. [1 ]Zhao, Xiaoxue, Weinan Zhang, and Jun Wang. "Interactive collaborative filtering." 
       Proceedings of the 22nd ACM international conference on Information & Knowledge Management. 2013.
    """
    def __init__(self, *args, **kwargs):
        """__init__.

        Args:
            args:
            kwargs:
        """
        super().__init__(*args, **kwargs)

    def reset(self, observation):
        """reset.

        Args:
            observation: 
        """                
        train_dataset = observation
        super().reset(train_dataset)

    def action_estimates(self, candidate_actions):
        """action_estimates.

        Args:
            candidate_actions: (user id, candidate_items)
        
        Returns:
            numpy.ndarray:
        """
        uid = candidate_actions[0]
        candidate_items = candidate_actions[1]
        b = self.bs[uid]
        A = self.As[uid]

        mean = np.dot(np.linalg.inv(A), b)

        items_score = mean @ self.items_means[candidate_items].T
        return items_score, None

    def update(self, observation, action, reward, info):
        """update.

        Args:
            observation:
            action: (user id, item)
            reward (float): reward
            info: 
        """
        uid = action[0]
        item = action[1]
        additional_data = info
        return super().update(observation, action, reward, info)



from os.path import dirname, realpath, sep, pardir
import sys, os

# sys.path.append(dirname(dirname(realpath(__file__))))
from irec.CandidateActions import CandidateActions
from irec.CandidateAction import CandidateAction

import numpy as np
import scipy.sparse

from collections import defaultdict
import pickle
import json

from typing import Any, Tuple


class ValueFunction:
    """ValueFunction.
    One of the goals in RS is to estimate the usefulness of a recommendation for a user. 
    From an RL perspective, in addition to an estimate of the immediate utility/reward of 
    an action, the agent aims to learn from experience the long-term value of an action. 
    The action value function Q(s, a) defines the value of performing an action a in state s.
    """

    def __init__(self, *args, **kwargs):
        """__init__.

        Args:
            args:
            kwargs:
        """
        super().__init__(*args, **kwargs)

    def reset(self, observation: Any):
        """reset.

        Reset all model attributes.

        Args:
            observation (Any): observation
        """
        train_dataset = observation
        pass
        # super().reset(train_dataset)

    def action_estimates(self, candidate_actions: CandidateActions) -> Tuple[Any, dict]:
        """action_estimates.
        
        Received as input the currently available actions the agent can perform and returns 
        the necessary information, limited to the actions the agent can perform at time step t,
        to allow a specific next action to be performed.

        Args:
            candidate_actions (CandidateActions): candidate_actions

        Returns:
            Tuple[Any, dict]: Scores and information
        """

        # uid = candidate_actions[0]
        # candidate_items = candidate_actions[1]
        return None, None

    def update(
        self, observation: Any, action: CandidateAction, reward: float, info: dict
    ) -> None:
        """update.
        
        According to the observations retrieved at each time step, the agent updates its 
        estimate of the value function based on the reward received.

        Args:
            observation (Any): observation
            action (CandidateAction): action
            reward (float): reward
            info (dict): info
        """
        # uid = action[0]
        # item = action[1]
        # additional_data = info
        return None



from .MF import *
from .ICFPMF import *
from .PMF import *
from .SVD import *
from .NMF import *
from .ICFPMFS import *
from .SVDPlusPlus import *

MF_MODELS = {}
for mf_model_class in [ICFPMFS, PMF, SVDPlusPlus, SVD, NMF, ICFPMF]:
    MF_MODELS[mf_model_class.__name__] = mf_model_class



import numpy as np
from tqdm import tqdm
from .ValueFunction import ValueFunction
from .ExperimentalValueFunction import ExperimentalValueFunction
import os
import random
import scipy.stats
from collections import defaultdict


class UCB(ExperimentalValueFunction):
    """Upper Confidence Bound.
    
    It is the original UCB that calculates a confidence interval for each item
    at each iteration and tries to shrink the confidence bounds [1]_.

    References
    ----------
    .. [1] Auer, P., Cesa-Bianchi, N. & Fischer, P. Finite-time Analysis of the 
       Multiarmed Bandit Problem. Machine Learning 47, 235–256 (2002).
    """

    def __init__(self, c, *args, **kwargs):
        """__init__.

        Args:
            args:
            kwargs:
            c:
        """

        super().__init__(*args, **kwargs)
        self.c = c

    def reset(self, observation):
        """reset.

        Args:
            observation: 
        """
        train_dataset = observation
        super().reset(train_dataset)
        self.train_dataset = train_dataset
        self.train_consumption_matrix = scipy.sparse.csr_matrix(
            (
                self.train_dataset.data[:, 2],
                (self.train_dataset.data[:, 0], self.train_dataset.data[:, 1]),
            ),
            (self.train_dataset.num_total_users, self.train_dataset.num_total_items),
        )
        self.num_total_items = self.train_dataset.num_total_items

        self.items_mean_values = np.zeros(self.num_total_items, dtype=np.float128)
        self.items_count = np.zeros(self.num_total_items, dtype=int)

        self.t = 1
        self.recent_predict = True
        for i in range(self.train_dataset.data.shape[0]):
            uid = int(self.train_dataset.data[i, 0])
            item = int(self.train_dataset.data[i, 1])
            reward = self.train_dataset.data[i, 2]
            # self.update(uid,item,reward,None)
            self.update(None, (uid, item), reward, None)

    def action_estimates(self, candidate_actions):
        """action_estimates.

        Args:
            candidate_actions: (user id, candidate_items)

        Returns:
            numpy.ndarray:
        """

        uid = candidate_actions[0]
        candidate_items = candidate_actions[1]
        with np.errstate(divide="ignore"):
            items_uncertainty = self.c * np.sqrt(
                2 * np.log(self.t) / self.items_count[candidate_items]
            )
            items_score = self.items_mean_values[candidate_items] + items_uncertainty
        self.recent_predict = True
        return items_score, None

    def update(self, observation, action, reward, info):
        """update.

        Args:
            observation:
            action: (user id, item)
            reward (float): reward
            info: 
        """
        uid = action[0]
        item = action[1]
        additional_data = info
        item = int(item)
        self.items_mean_values[item] = (
            self.items_mean_values[item] * self.items_count[item] + reward
        ) / (self.items_count[item] + 1)
        self.items_count[item] += 1
        if self.recent_predict:
            self.t += 1
            self.recent_predict = False



import numpy as np
from tqdm import tqdm
from . import ValueFunction, Entropy, Entropy0, MostPopular, LogPopEnt, ExperimentalValueFunction
import matplotlib.pyplot as plt
import scipy.stats
import os
import random
import metrics


class DistinctPopular(ExperimentalValueFunction):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def reset(self, observation):
        train_dataset = observation
        super().reset(train_dataset)
        self.train_dataset = train_dataset
        self.train_consumption_matrix = scipy.sparse.csr_matrix(
            (self.train_dataset.data[:, 2],
             (self.train_dataset.data[:, 0], self.train_dataset.data[:, 1])),
            (self.train_dataset.num_total_users,
             self.train_dataset.num_total_items))
        self.num_total_items = self.train_dataset.num_total_items

        self.items_entropy = Entropy.get_items_entropy(
            self.train_consumption_matrix)
        np.seterr('warn')
        self.items_popularity = MostPopular.get_items_popularity(
            self.train_consumption_matrix, normalize=False)
        self.items_distance = metrics.get_items_distance(
            self.train_consumption_matrix)

        self.top_iids = defaultdict([])
        # num_total_items = self.train_consumption_matrix.shape[1]

    def action_estimates(self, candidate_actions):
        uid = candidate_actions[0]
        candidate_items = candidate_actions[1]
        if len(self.top_iids) > 0:
            candidate_items_distance = np.mean(
                self.items_distance[self.top_iids[uid]][:, candidate_items],
                axis=0)
        else:
            candidate_items_distance = 1

        return candidate_items_distance * items_popularity[cadidate_items], None

    def update(self, observation, action, reward, info):
        uid = action[0]
        item = action[1]
        additional_data = info
        self.top_iids[uid].append(item)

        # correlation = scipy.stats.pearsonr(items_entropy,items_popularity)[0]

        # fig, ax = plt.subplots()
        # ax.scatter(items_entropy,items_popularity,marker="D",color='darkblue')
        # ax.set_ylabel("Popularity")
        # ax.set_xlabel("Entropy")
        # ax.text(0.3, 0.9 , f'Correlation coefficient: {correlation:.2f}', color='k',
        #         ha='center', va='center',
        #         bbox=dict(facecolor='none', edgecolor='k', pad=10.0),
        #         transform = ax.transAxes)
        # for start, end, color in [(0,10,'green'),(10,20,'red'),(20,30,'darkred'),(30,40,'yellow'),(40,50,'orange')]:
        #     ax.scatter(items_entropy[top_iids[start:end]],items_popularity[top_iids[start:end]],marker='D',color=color)
        #     print("[%d,%d] sum(popularity)=%.2f sum(entropy)=%.2f"%(start,end,
        #                                                          np.sum(items_popularity[top_iids[start:end]]/np.max(items_popularity)),
        #                                                          np.sum(items_entropy[top_iids[start:end]]/np.max(items_entropy))))
        # fig.savefig(os.path.join(self.DIRS['img'],"corr_popent_"+self.get_id()+".png"))

        # num_total_users = len(uids)
        # for idx_uid in tqdm(range(num_total_users)):
        #     uid = uids[idx_uid]
        #     self.results[uid].extend(top_iids)
        # self.save_results()



import numpy as np

import numbers
import copy
from numpy.lib.npyio import save
from tqdm import tqdm, trange
from . import ValueFunction, ExperimentalValueFunction
import os
import random
import scipy.stats
from collections import defaultdict

def _create_attribute_with_check(value):
    if isinstance(value,numbers.Number):
        return defaultdict(lambda: value)
    elif isinstance(value,dict):
        return defaultdict(lambda:1,value)

class GenericThompsonSampling(ExperimentalValueFunction):

    def __init__(self, alpha_0, beta_0, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.alpha_0 = alpha_0
        self.beta_0 = beta_0
        #

    def reset(self, observation):
        train_dataset = observation
        super().reset(train_dataset)
        # self.train_dataset = train_dataset
        # self.train_consumption_matrix = scipy.sparse.csr_matrix(
        # (self.train_dataset.data[:, 2],
        # (self.train_dataset.data[:, 0], self.train_dataset.data[:, 1])),
        # (self.train_dataset.num_total_users,
        # self.num_total_items = self.train_dataset.num_total_items

        self.alphas = _create_attribute_with_check(self.alpha_0)
        self.betas = _create_attribute_with_check(self.beta_0)

        # self.train_dataset.num_total_items))
        # for i in range(self.train_dataset.data.shape[0]):
            # uid = int(self.train_dataset.data[i, 0])
            # item = int(self.train_dataset.data[i, 1])
            # reward = self.train_dataset.data[i, 2]
            # # self.action_estimates()
            # # self.update(uid, item, reward, None)
            # self.update(None, (uid, item), reward, None)

    def action_estimates(self, candidate_actions):
        items_score = np.random.beta([self.alphas[ca] for ca in candidate_actions],
                                     [self.betas[ca] for ca in candidate_actions])
        # info = {'a': copy.copy(self.alphas), 'b': copy.copy(self.betas)}
        return items_score, None
        # return items_score, info

    def update(self, observation, action, reward, info):
        # additional_data = info
        reward = 1 if (reward >= 4) else 0
        self.alphas[action] += reward
        self.betas[action] += 1 - reward



import ctypes
from collections import defaultdict

import mf
import numpy as np
import scipy
import scipy.stats

from threadpoolctl import threadpool_limits
from tqdm import tqdm

import value_functions

from .ExperimentalValueFunction import ExperimentalValueFunction
from .MFValueFunction import MFValueFunction


def _prediction_rule(A, b, items_weights, alpha):
    mean = np.dot(np.linalg.inv(A), b)
    items_uncertainty = alpha * np.sqrt(
        np.sum(items_weights.dot(np.linalg.inv(A)) * items_weights, axis=1)
    )
    items_user_similarity = mean @ items_weights.T
    items_score = items_user_similarity + items_uncertainty
    return items_score


class LinUCB(MFValueFunction):
    def __init__(self, alpha, zeta=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if alpha != None:
            self.alpha = alpha
        elif zeta != None:
            self.alpha = 1 + np.sqrt(np.log(2 / zeta) / 2)

    def reset(self, observation):
        train_dataset = observation
        super().reset(train_dataset)
        self.train_dataset = train_dataset
        self.train_consumption_matrix = scipy.sparse.csr_matrix(
            (
                self.train_dataset.data[:, 2],
                (self.train_dataset.data[:, 0], self.train_dataset.data[:, 1]),
            ),
            (self.train_dataset.num_total_users, self.train_dataset.num_total_items),
        )
        self.num_total_items = self.train_dataset.num_total_items

        mf_model = mf.SVD(num_lat=self.num_lat)
        mf_model.fit(self.train_consumption_matrix)
        self.items_weights = mf_model.items_weights
        self.num_latent_factors = len(self.items_weights[0])

        self.I = np.eye(len(self.items_weights[0]))
        self.bs = defaultdict(lambda: np.ones(self.num_latent_factors))

        self.As = defaultdict(lambda: self.I.copy())

    def action_estimates(self, candidate_actions):
        uid = candidate_actions[0]
        candidate_items = candidate_actions[1]
        b = self.bs[uid]
        A = self.As[uid]
        items_score = _prediction_rule(
            A, b, self.items_weights[candidate_items], self.alpha
        )
        return items_score, None

    def update(self, observation, action, reward, info):
        uid = action[0]
        item = action[1]
        additional_data = info
        max_item_latent_factors = self.items_weights[item]
        b = self.bs[uid]
        A = self.As[uid]
        A += max_item_latent_factors[:, None].dot(max_item_latent_factors[None, :])
        b += reward * max_item_latent_factors



from . import MF
import scipy.sparse
import numpy as np


class SVD(MF):
    def fit(self, training_matrix):
        u, s, vt = scipy.sparse.linalg.svds(
            scipy.sparse.csr_matrix(training_matrix), k=self.num_lat)
        self.users_weights = u
        self.items_weights = vt.T * s



import numpy as np
import scipy
import scipy.stats
import sys, os
import random
from threadpoolctl import threadpool_limits
import ctypes
import collections
import metrics
from tqdm import tqdm

from irec.utils.utils import run_parallel
from . import MF
from numba import jit


@jit(nopython=True)
def _multivariate_normal(x, mu=0, sigma=1):
    """focus in speed and readability multivariate_normal, no checks like scipy and others libraries
be sure to pass float array and matrix
very limited to be more fast"""
    k = len(x)
    return (1/(np.sqrt(2*np.pi)**k * np.linalg.det(sigma)))*\
        np.exp((-1/2)*(x-mu) @ np.linalg.inv(sigma)@(x - mu))


@jit(nopython=True)
def _apply_multivariate_normal(xs, mu=0, sigma=1):
    n = len(xs)
    result = np.zeros(n)
    for i in range(n):
        result[i] = _multivariate_normal(xs[i], mu, sigma)
    return result


def _norm_sum_probabilities(x):
    return np.sum(-np.log(x))
    # return scipy.special.logsumexp(x)


def _norm_ratings(x, highest_value, lowest_value):
    return 2 * (x - lowest_value) / (highest_value - lowest_value) - 1


def _unnorm_ratings(x, highest_value, lowest_value):
    return (x + 1) / 2 * (highest_value - lowest_value) + lowest_value


class ICFPMFS(MF):
    def __init__(self,
                 iterations=20,
                 var=0.05,
                 user_var=0.01,
                 item_var=0.01,
                 stop_criteria=0.0009,
                 *args,
                 **kwargs):
        super().__init__(*args, **kwargs)
        self.iterations = iterations
        self.var = var
        self.user_var = user_var
        self.item_var = item_var
        self.stop_criteria = stop_criteria

    def get_user_lambda(self):
        return self.var / self.user_var

    def get_item_lambda(self):
        return self.var / self.item_var

    def load_var(self, training_matrix):
        decimals = 4
        # self.var = np.mean(training_matrix.data**2) - np.mean(training_matrix.data)**2
        # self.user_var = np.mean([np.mean(i.data**2) - np.mean(i.data)**2 if i.getnnz()>0 else 0 for i in training_matrix])
        # self.item_var = np.mean([np.mean(i.data**2) - np.mean(i.data)**2 if i.getnnz()>0 else 0 for i in training_matrix.transpose()])

        # self.var = np.round(self.var,decimals)
        # self.user_var = np.round(self.user_var,decimals)
        # self.item_var = np.round(self.item_var,decimals)

    def fit(self, training_matrix):
        super().fit()
        train_uids = np.unique(training_matrix.tocoo().row)
        self.objective_values = []
        self.best = None
        decimals = 4
        self.user_lambda = self.var / self.user_var
        self.item_lambda = self.var / self.item_var
        self.r_mean = np.mean(training_matrix.data)
        # self.r_mean = np.
        self_id = id(self)
        # training_matrix = self.normalize_matrix(training_matrix)

        self.training_matrix = training_matrix
        num_users = training_matrix.shape[0]
        num_items = training_matrix.shape[1]
        self.lowest_value = lowest_value = np.min(training_matrix)
        self.highest_value = highest_value = np.max(training_matrix)
        self.observed_ui = observed_ui = (training_matrix.tocoo().row,
                                          training_matrix.tocoo().col
                                          )  # itens observed by some user
        self.I = I = np.eye(self.num_lat)
        self.users_weights = np.random.multivariate_normal(
            np.zeros(self.num_lat), self.user_var * I,
            training_matrix.shape[0])
        self.items_weights = np.random.multivariate_normal(
            np.zeros(self.num_lat), self.item_var * I,
            training_matrix.shape[1])
        # self.users_weights[~np.isin(list(range(self.users_weights.shape[0])), train_uids)] = np.ones(self.num_lat)

        # self.users_weights = np.mean(training_matrix.data)*np.random.rand(num_users,self.num_lat)
        # self.items_weights = np.mean(training_matrix.data)*np.random.rand(num_items,self.num_lat)
        # self.users_weights = 0.1*np.random.multivariate_normal(np.zeros(self.num_lat),self.user_var*I,training_matrix.shape[0])
        # self.items_weights = 0.1*np.random.multivariate_normal(np.zeros(self.num_lat),self.item_var*I,training_matrix.shape[1])

        self.users_observed_items = collections.defaultdict(list)
        self.items_observed_users = collections.defaultdict(list)
        self.users_observed_items_ratings = collections.defaultdict(list)
        self.items_observed_users_ratings = collections.defaultdict(list)
        for uid, iid in zip(*self.observed_ui):
            self.users_observed_items[uid].append(iid)
            self.items_observed_users[iid].append(uid)

        for uid, iids in self.users_observed_items.items():
            self.users_observed_items_ratings[uid] = training_matrix[uid,
                                                                     iids].data

        for iid, uids in self.items_observed_users.items():
            self.items_observed_users_ratings[iid] = training_matrix[uids,
                                                                     iid].data

        best_objective_value = None

        last_objective_value = None

        # users_probabilities_args = [(self_id, i) for i in range(self.users_weights.shape[0])]
        # items_probabilities_args = [(self_id, i) for i in range(self.items_weights.shape[0])]

        # without burning
        np.seterr('warn')
        tq = tqdm(range(self.iterations))
        for i in tq:
            with threadpool_limits(limits=1, user_api='blas'):
                # for to_run in random.sample([1,2],2):
                for to_run in [1, 2]:
                    if to_run == 1:
                        self.users_means = np.zeros((num_users, self.num_lat))
                        self.users_covs = np.zeros(
                            (num_users, self.num_lat, self.num_lat))
                        args = [(
                            self_id,
                            i,
                        ) for i in train_uids]
                        results = run_parallel(self.compute_user_weight,
                                               args,
                                               use_tqdm=False)
                        for uid, (mean, cov,
                                  weight) in zip(train_uids, results):
                            self.users_means[uid] = mean
                            self.users_covs[uid] = cov
                            self.users_weights[uid] = weight
                    else:
                        self.items_means = np.zeros((num_items, self.num_lat))
                        self.items_covs = np.zeros(
                            (num_items, self.num_lat, self.num_lat))
                        args = [(
                            self_id,
                            i,
                        ) for i in range(num_items)]
                        results = run_parallel(self.compute_item_weight,
                                               args,
                                               use_tqdm=False)
                        for iid, (mean, cov, weight) in enumerate(results):
                            self.items_means[iid] = mean
                            self.items_covs[iid] = cov
                            self.items_weights[iid] = weight

            predicted = self.predict(observed_ui)

            tq.set_description('rmse={:.3f}'.format(
                metrics.rmse(
                    training_matrix.data,
                    _unnorm_ratings(predicted, self.lowest_value,
                                    self.highest_value))))
            tq.refresh()

        #     # objective_value = _norm_sum_probabilities(scipy.stats.norm.pdf(training_matrix.data,predicted,self.var))\
        #     #     + _norm_sum_probabilities(_apply_multivariate_normal(self.users_weights,np.zeros(self.num_lat),self.var*self.I))\
        #     #     + _norm_sum_probabilities(_apply_multivariate_normal(self.items_weights,np.zeros(self.num_lat),self.var*self.I))

        #     objective_value = np.sum((training_matrix.data - predicted)**2)/2 +\
        #         self.user_lambda/2 * np.sum(np.linalg.norm(self.users_weights,axis=1)**2) +\
        #         self.item_lambda/2 * np.sum(np.linalg.norm(self.items_weights,axis=1)**2)

        #     self.objective_values.append(objective_value)

        #     if self.best == None:
        #         self.best = self.__deepcopy__()
        #         best_objective_value = objective_value
        #     else:
        #         if objective_value < best_objective_value:
        #             self.best = self.__deepcopy__()
        #             best_objective_value = objective_value

        #     tq.set_description('cur={:.3f},best={:.3f}'.format(objective_value,best_objective_value))
        #     tq.refresh()

        #     # predicted = self.predict(observed_ui)
        #     # rmse=metrics.rmse(training_matrix.data,predicted)
        #     # objective_value = rmse
        #     # print("RMSE",rmse)
        #     # if np.fabs(objective_value - last_objective_value) <= self.stop_criteria:
        #     #     self.objective_value = objective_value
        #     #     print("Achieved convergence with %d iterations"%(i+1))
        #     #     break
        #     # last_objective_value = objective_value

        #     # sparse_predicted = self.get_sparse_predicted(observed_ui_pair)
        #     # rmse=np.sqrt(np.mean((sparse_predicted - training_matrix.data)**2))
        #     # objective_value = np.sum((training_matrix.data - sparse_predicted)**2)/2 +\
        #     #     self.user_lambda/2 * np.sum(np.linalg.norm(self.users_weights,axis=1)**2) +\
        #     #     self.item_lambda/2 * np.sum(np.linalg.norm(self.items_weights,axis=1)**2)
        #     # print("Objective value",objective_value)
        #     # #     self.objective_values.append(objective_value)
        #     # print("RMSE",rmse)
        # self.__dict__.update(self.best.__dict__)
        # del self.best
        del self.r_mean
        del self.user_lambda
        del self.item_lambda
        del self.users_observed_items
        del self.users_observed_items_ratings
        del self.items_observed_users
        del self.items_observed_users_ratings
        del self.observed_ui
        del self.I
        del self.training_matrix
        del self.lowest_value
        del self.highest_value

    @staticmethod
    def _user_probability(obj_id, uid):
        self = ctypes.cast(obj_id, ctypes.py_object).value
        return scipy.stats.multivariate_normal.pdf(self.users_weights[uid],
                                                   np.zeros(self.num_lat),
                                                   self.var * self.I)

    @staticmethod
    def _item_probability(obj_id, iid):
        self = ctypes.cast(obj_id, ctypes.py_object).value
        return scipy.stats.multivariate_normal.pdf(self.items_weights[iid],
                                                   np.zeros(self.num_lat),
                                                   self.var * self.I)

    @staticmethod
    def compute_user_weight(obj_id, uid):
        self = ctypes.cast(obj_id, ctypes.py_object).value
        training_matrix = self.training_matrix
        lowest_value = self.lowest_value
        I = self.I
        observed = self.users_observed_items[uid]
        tmp = np.linalg.inv(
            (np.dot(self.items_weights[observed].T,
                    self.items_weights[observed]) + I * self.user_lambda))
        mean = tmp.dot(self.items_weights[observed].T).dot(
            _norm_ratings(self.users_observed_items_ratings[uid],
                          self.lowest_value, self.highest_value))
        cov = tmp * self.var
        return mean, cov, np.random.multivariate_normal(mean, cov)
        # return mean, cov, scipy.stats.multivariate_normal.pdf(self.users_weights[uid],mean,cov)
        # return mean, cov, scipy.stats.multivariate_normal.pdf(np.random.multivariate_normal(mean,cov),mean,cov)

    @staticmethod
    def compute_item_weight(obj_id, iid):
        self = ctypes.cast(obj_id, ctypes.py_object).value
        training_matrix = self.training_matrix
        lowest_value = self.lowest_value
        I = self.I
        observed = self.items_observed_users[iid]
        tmp = np.linalg.inv(
            (np.dot(self.users_weights[observed].T,
                    self.users_weights[observed]) + I * self.item_lambda))
        mean = tmp.dot(self.users_weights[observed].T).dot(
            _norm_ratings(self.items_observed_users_ratings[iid],
                          self.lowest_value, self.highest_value))
        cov = tmp * self.var
        return mean, cov, np.random.multivariate_normal(mean, cov)
        # return mean, cov, scipy.stats.multivariate_normal.pdf(self.items_weights[iid],mean,cov)
        # return mean, cov, scipy.stats.multivariate_normal.pdf(np.random.multivariate_normal(mean,cov),mean,cov)

    def __deepcopy__(self):
        new = type(self)()
        new.__dict__.update(self.__dict__)
        new.users_weights = self.users_weights.copy()
        new.users_means = self.users_means.copy()
        new.users_covs = self.users_covs.copy()
        new.items_weights = self.items_weights.copy()
        new.items_means = self.items_means.copy()
        new.items_covs = self.items_covs.copy()
        return new



from . import DatasetLoader
import irec.utils.dataset as dataset


class DatasetLoaderFactory:
    def create(self, dataset_name, dataset_parameters):
        dl = None
        if dataset_name == "MovieLens 100k O":
            dl = DatasetLoader.ML100kDatasetLoader(**dataset_parameters)
        elif dataset_name in [
            "MovieLens 1M",
            "MovieLens 10M",
            "MovieLens 20M",
            "LastFM 5k",
            "Kindle Store",
            "Kindle 4k",
            "Netflix 10k",
            "Good Books",
            "Yahoo Music",
            "Good Reads 10k",
            "Yahoo Music 5k",
            "Yahoo Music 10k",
        ]:
            dl = DatasetLoader.DefaultDatasetLoader(
                dataset.DefaultDataset(), **dataset_parameters
            )
        elif dataset_name in [
            "MovieLens 1M Validation",
            "MovieLens 10M Validation",
            "MovieLens 20M Validation",
            "LastFM 5k Validation",
            "Kindle Store Validation",
            "Kindle 4k Validation",
            "Good Books Validation",
            "Yahoo Music Validation",
            "Netflix 10k Validation",
            "Good Reads 10k Validation",
            "Yahoo Music 5k Validation",
            "Yahoo Music 10k Validation",
        ]:
            dl = DatasetLoader.DefaultValidationDatasetLoader(
                dataset.DefaultDataset(), **dataset_parameters
            )
        elif dataset_name == "Netflix":
            dl = DatasetLoader.DefaultDatasetLoader(
                dataset.Netflix(), **dataset_parameters
            )
        elif dataset_name in ["Netflix Validation"]:
            dl = DatasetLoader.DefaultDatasetLoader(
                dataset.Netflix(), **dataset_parameters
            )
        elif dataset_name in [
            "Kindle 4k TRTE",
            "Good Books TRTE",
            "LastFM 5k TRTE",
            "MovieLens 10M TRTE",
            "MovieLens 100k TRTE",
            "Yahoo Music TRTE",
            "GoodBooks TRTE",
            "LastFM 5k TRTE",
            "Netflix 10k TRTE",
        ]:
            dl = DatasetLoader.TRTEDatasetLoader(**dataset_parameters)
        elif dataset_name in [
            "Kindle 4k TRTE Validation",
            "Good Books TRTE Validation",
            "LastFM 5k TRTE Validation",
            "GoodBooks TRTE Validation",
            "Yahoo Music TRTE Validation",
            "LastFM 5k TRTE Validation",
            "MovieLens 10M TRTE Validation",
            "Netflix 10k TRTE Validation",
        ]:
            dl = DatasetLoader.TRTEValidationDatasetLoader(**dataset_parameters)
        else:
            raise IndexError(dataset_name)
        return dl



import numpy as np
from tqdm import tqdm
from . import ValueFunction, Entropy, MostPopular, LogPopEnt, ExperimentalValueFunction
import matplotlib.pyplot as plt
import scipy.stats
import os


class LogPopEnt(ExperimentalValueFunction.ExperimentalValueFunction):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    @staticmethod
    def get_items_logpopent(items_popularity, items_entropy, k=None):
        if k is not None:
            items_logpopent = (items_entropy**
                               k) * np.ma.log(items_popularity).filled(0)**(1 -
                                                                            k)
        else:
            items_logpopent = items_entropy * np.ma.log(
                items_popularity).filled(0)
        return np.dot(items_logpopent, 1 / np.max(items_logpopent))

    def reset(self, observation):
        train_dataset = observation
        super().reset(train_dataset)
        self.train_dataset = train_dataset
        self.train_consumption_matrix = scipy.sparse.csr_matrix(
            (self.train_dataset.data[:, 2],
             (self.train_dataset.data[:, 0], self.train_dataset.data[:, 1])),
            (self.train_dataset.num_total_users,
             self.train_dataset.num_total_items))

        items_entropy = Entropy.get_items_entropy(
            self.train_consumption_matrix)
        items_popularity = MostPopular.get_items_popularity(
            self.train_consumption_matrix, normalize=False)
        self.items_logpopent = LogPopEnt.get_items_logpopent(
            items_popularity, items_entropy)

    def action_estimates(self, candidate_actions):
        uid = candidate_actions[0]
        candidate_items = candidate_actions[1]
        items_score = self.items_logpopent[candidate_items]
        return items_score, None

    def update(self, observation, action, reward, info):
        uid = action[0]
        item = action[1]
        additional_data = info
        pass



from irec.CandidateActions import CandidateActions
from sklearn.cluster import KMeans
import numpy as np
from collections import defaultdict
import irec.value_functions.Entropy

from value_functions.MostPopular import MostPopular
from value_functions.Entropy import Entropy
import scipy.sparse
from value_functions.LogPopEnt import LogPopEnt
import random
import mf


class ActionSelectionPolicy:
    def __init__(self, *args, **kwargs):
        pass

    def select_actions(self, actions: CandidateActions, action_estimates, actions_num):
        raise NotImplementedError

    def update(self, observation, action, reward, info):
        raise NotImplementedError

    def reset(self, observation):
        raise NotImplementedError


class ASPGreedy(ActionSelectionPolicy):
    def select_actions(self, actions, action_estimates, actions_num):
        return (
            actions[0],
            actions[1][np.argpartition(action_estimates, -actions_num)[-actions_num:]],
        ), None

    def update(self, observation, action, reward, info):
        pass

    def reset(self, observation):
        pass


class ASPEGreedy(ActionSelectionPolicy):
    def __init__(self, epsilon, *args, **kwargs):
        super().__init__(self, *args, **kwargs)
        self.epsilon = epsilon

    def select_actions(self, actions, action_estimates, actions_num):
        greedy_actions = np.argpartition(action_estimates, -actions_num)[-actions_num:][
            ::-1
        ]
        actions_indexes = []
        for i in range(actions_num):
            if self.epsilon < np.random.rand():
                j = 0
                while True:
                    action_index = greedy_actions[j]
                    if action_index not in actions_indexes:
                        break
                    j += 1
                actions_indexes.append(action_index)
            else:
                while True:
                    action_index = np.random.randint(len(action_estimates))
                    if action_index not in actions_indexes:
                        break
                actions_indexes.append(action_index)

        return (actions[0], actions[1][actions_indexes]), None

    def update(self, observation, action, reward, info):
        pass

    def reset(self, observation):
        pass


class ASPReranker(ActionSelectionPolicy):
    def __init__(self, rule, input_filter_size, rerank_limit, *args, **kwargs):
        super().__init__(self, *args, **kwargs)
        self.rule = rule
        self.input_filter_size = input_filter_size
        self.rerank_limit = rerank_limit

    def select_actions(self, actions, action_estimates, actions_num):
        if self.users_num_consumption[actions[0]] > self.rerank_limit:
            return (
                actions[0],
                actions[1][
                    np.argpartition(action_estimates, -actions_num)[-actions_num:]
                ],
            ), None

        top_estimates_index_actions = np.argpartition(
            action_estimates, -self.input_filter_size
        )[-self.input_filter_size :][::-1]

        actions = (actions[0], actions[1][top_estimates_index_actions])
        action_estimates = action_estimates[top_estimates_index_actions]

        rule_action_estimates = self.rule.action_estimates(actions)[0]
        top_rule_index_actions = np.argpartition(rule_action_estimates, -actions_num)[
            -actions_num:
        ][::-1]
        return (actions[0], actions[1][top_rule_index_actions]), None

    def update(self, observation, action, reward, info):
        if reward >= 4:
            self.users_num_consumption[action[0]] += 1
        self.rule.update(observation, action, reward, info)

    def reset(self, observation):
        self.users_num_consumption = defaultdict(int)
        self.rule.reset(observation)


class ASPGenericGreedy(ActionSelectionPolicy):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        # self.save_played_actions = save_played_actions
        pass

    def select_actions(self, actions, action_estimates, actions_num):
        # print(actions,action_estimates)
        # info = {''}
        actions = actions[
            np.argpartition(action_estimates, -actions_num)[-actions_num:]
        ]
        # if self.save_played_actions:
        # info = {'played_actions': factions}
        return actions, None

    def update(self, observation, action, reward, info):
        pass

    def reset(self, observation):
        pass


class ASPICGreedy(ActionSelectionPolicy):
    def __init__(self, stop, num_clusters, num_lat, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.stop = stop
        self.num_clusters = num_clusters
        self.num_lat = num_lat

    def choose_group(self, uid):
        b = self.users_num_consumption_groups[uid]
        return np.random.choice(np.flatnonzero(b == b.min()))

    def get_item_group(self, item):
        return self.groups_items[item]

    def select_actions(self, actions, action_estimates, actions_num):
        uid = actions[0]
        if self.users_num_consumption[uid] >= self.stop:
            return (
                actions[0],
                actions[1][
                    np.argpartition(action_estimates, -actions_num)[-actions_num:]
                ],
            ), None
        else:
            i=0
            while True:
                i+=1
                # print(i)
                # g = self.choose_group(uid)
                # try:
                g= random.randint(0,self.num_clusters)
                g_items = np.array(self.groups_items[g])
                # print(actions[1],g_items)
                g_items = np.intersect1d(actions[1],g_items)
                # print(g_items)
                if len(g_items) == 0:
                    continue
                g_items = list(g_items)
                top_item = g_items[np.argmax(self.items_popularity[g_items])]
                break
                    # break
                # except:
                    # continue

            return (
                actions[0],
                [top_item],
            ), None

    def update(self, observation, action, reward, info):
        uid = action[0]
        iid = action[1]
        self.users_num_consumption[uid] += 1
        pass

    def reset(self, observation):
        train_dataset = observation
        self.train_dataset = train_dataset

        self.train_consumption_matrix = scipy.sparse.csr_matrix(
            (
                self.train_dataset.data[:, 2],
                (self.train_dataset.data[:, 0], self.train_dataset.data[:, 1]),
            ),
            (self.train_dataset.num_total_users, self.train_dataset.num_total_items),
        )
        mf_model = mf.SVD(num_lat=self.num_lat)
        mf_model.fit(self.train_consumption_matrix)
        self.items_weights = mf_model.items_weights
        self.num_latent_factors = len(self.items_weights[0])

        kmeans = KMeans(self.num_clusters).fit(self.items_weights)

        self.num_total_items = self.train_dataset.num_total_items
        self.num_total_users = self.train_dataset.num_total_users

        self.groups_items = defaultdict(list)
        print(len(kmeans.labels_))
        for uid, group in enumerate(kmeans.labels_):
            self.groups_items[group].append(uid)

        self.groups = list(range(self.num_clusters))
        self.users_num_consumption = defaultdict(int)
        self.users_num_consumption_groups = defaultdict(
            lambda: np.zeros(self.num_clusters)
        )

        self.items_entropy = Entropy.get_items_entropy(self.train_consumption_matrix)
        self.items_popularity = MostPopular.get_items_popularity(
            self.train_consumption_matrix, normalize=False
        )
        self.items_logpopent = LogPopEnt.get_items_logpopent(
            self.items_popularity, self.items_entropy
        )



import numpy as np
from tqdm import tqdm
#import util
from threadpoolctl import threadpool_limits
import ctypes
import scipy.spatial
import matplotlib.pyplot as plt
import os
import pickle
import sklearn
import scipy.optimize
import scipy
import mf
from collections import defaultdict
from .MFValueFunction import MFValueFunction
import value_functions
import value_functions.Entropy
import value_functions.MostPopular
import value_functions.LogPopEnt


def _prediction_rule(A, b, items_weights, alpha):
    user_latent_factors = np.dot(np.linalg.inv(A), b)
    items_uncertainty = np.sqrt(
        np.sum(items_weights.dot(np.linalg.inv(A)) * items_weights, axis=1))
    items_user_similarity = user_latent_factors @ items_weights.T
    user_model_items_score = items_user_similarity + alpha * items_uncertainty
    return user_model_items_score


class OurMethod2var(MFValueFunction):
    def __init__(self, alpha, lambda_u, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.alpha = alpha
        self.lambda_u = lambda_u


    def reset(self, observation):
        train_dataset = observation
        super().reset(train_dataset)
        self.train_dataset = train_dataset
        self.train_consumption_matrix = scipy.sparse.csr_matrix(
            (self.train_dataset.data[:, 2],
             (self.train_dataset.data[:, 0], self.train_dataset.data[:, 1])),
            (self.train_dataset.num_total_users,
             self.train_dataset.num_total_items))
        self.num_total_items = self.train_dataset.num_total_items

        mf_model = mf.SVD(num_lat=self.num_lat)
        mf_model.fit(self.train_consumption_matrix)
        self.items_weights = mf_model.items_weights
        self.num_latent_factors = len(self.items_weights[0])

        items_entropy = value_functions.Entropy.Entropy.get_items_entropy(
            self.train_consumption_matrix)
        items_popularity = value_functions.MostPopular.MostPopular.get_items_popularity(
            self.train_consumption_matrix, normalize=False)
        # self.items_bias = value_functions.PPELPE.get_items_ppelpe(items_popularity,items_entropy)
        self.items_bias = value_functions.LogPopEnt.LogPopEnt.get_items_logpopent(
            items_popularity, items_entropy)
        print(self.items_bias.min(), self.items_bias.max())
        assert (self.items_bias.min() >= 0
                and np.isclose(self.items_bias.max(), 1))

        # regression_model = sklearn.linear_model.LinearRegression()
        res = scipy.optimize.minimize(
            lambda x, items_weights, items_bias: np.sum(
                (items_bias - x @ items_weights.T)**2),
            np.ones(self.num_latent_factors),
            args=(self.items_weights, self.items_bias),
            method='BFGS',
        )
        self.initial_b = res.x

        print(
            np.corrcoef(self.items_bias,
                        self.initial_b @ self.items_weights.T)[0, 1])

        self.I = np.eye(len(self.items_weights[0]))
        self.bs = defaultdict(lambda: self.initial_b.copy())
        self.As = defaultdict(lambda: self.lambda_u * self.I.copy())
        # items_score = _prediction_rule(self.lambda_u*self.I,self.initial_b,self.items_weights,self.alpha)

        # print("WSCBvar items score correlation with popularity:",scipy.stats.pearsonr(items_score,items_popularity),self.train_dataset.num_total_users, self.train_dataset.num_total_items)
        # print("WSCBvar items score correlation with entropy:",scipy.stats.pearsonr(items_score,items_entropy),self.train_dataset.num_total_users, self.train_dataset.num_total_items)

    def action_estimates(self, candidate_actions):
        uid = candidate_actions[0]
        candidate_items = candidate_actions[1]
        b = self.bs[uid]
        A = self.As[uid]
        user_model_items_score = _prediction_rule(
            A, b, self.items_weights[candidate_items], self.alpha)
        items_score = user_model_items_score
        return items_score, None

    def update(self, observation, action, reward, info):
        uid = action[0]
        item = action[1]
        additional_data = info
        max_item_latent_factors = self.items_weights[item]
        b = self.bs[uid]
        A = self.As[uid]
        A += max_item_latent_factors[:, None].dot(
            max_item_latent_factors[None, :])
        b += reward * max_item_latent_factors
